#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements the package definition.
"""
from typing import Dict, List, Tuple, Union, Callable
import warnings
from warnings import warn
import os
import shutil
import json
import re
from collections import OrderedDict

import pandas as pd
import numpy as np

from ckanapi_harvesters.auxiliary.ckan_progress_callbacks import CkanProgressCallback, CkanProgressCallbackABC, CkanCallbackLevel
from ckanapi_harvesters.policies.data_format_policy_errors import DataPolicyError
from ckanapi_harvesters.policies.data_format_policy import CkanPackageDataFormatPolicy
from ckanapi_harvesters.ckan_api import CkanApi, CkanApiMap
from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig
from ckanapi_harvesters.auxiliary.ckan_model import CkanVisibility, CkanState, CkanPackageInfo, CkanResourceInfo, CkanLicenseInfo
from ckanapi_harvesters.auxiliary.path import sanitize_path, make_path_relative
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, assert_or_raise, find_duplicates
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.auxiliary.ckan_errors import (DuplicateNameError, ForbiddenNameError, MissingIdError,
                                                      MandatoryAttributeError, FileOrDirNotExistError)
from ckanapi_harvesters.auxiliary.ckan_configuration import unlock_external_url_resource_download, unlock_no_ca
from ckanapi_harvesters.builder.builder_errors import (UnsupportedBuilderVersionError,
                                                       MissingDataStoreColumnsSheet)
from ckanapi_harvesters.builder import BUILDER_FILE_FORMAT_VERSION as BUILDER_VER
from ckanapi_harvesters.builder.builder_resource import BuilderResourceABC, BuilderUrl
from ckanapi_harvesters.builder.builder_resource_multi_file import BuilderMultiFile, multi_file_exclude_other_files
from ckanapi_harvesters.builder.builder_resource_datastore_url import BuilderDataStoreUrl
from ckanapi_harvesters.builder.builder_resource_datastore import BuilderDataStoreABC
from ckanapi_harvesters.builder.builder_resource_multi_datastore import BuilderMultiDataStore
from ckanapi_harvesters.builder.builder_resource_datastore_multi_abc import BuilderDataStoreMultiABC
from ckanapi_harvesters.builder.builder_resource_datastore_multi_harvester import BuilderDataStoreHarvester
from ckanapi_harvesters.builder.builder_resource_init import init_resource_from_df, init_resource_from_ckan
from ckanapi_harvesters.builder.builder_ckan import BuilderCkan
from ckanapi_harvesters.auxiliary.external_code_import import PythonUserCode, unlock_external_code_execution
from ckanapi_harvesters.harvesters.file_formats.user_format import UserFileFormat

self_dir = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
example_package_xls = os.path.join(self_dir, "builder_package_example.xlsx")
example_package_resources_dir = os.path.join(self_dir, "example", "package")
builder_default_package_state:CkanState = CkanState.Draft
initial_package_building_state:Union[CkanState,None] = CkanState.Draft

def load_help_page_df(*, engine:str=None) -> pd.DataFrame:
    with open(example_package_xls, "rb") as stream:  # read-only mode
        with pd.ExcelFile(stream, engine=engine) as help_file:
            help_df = pd.read_excel(help_file, sheet_name="help", header=None)
            help_file.close()
    return help_df

forbidden_resource_names = {"ckan", "info", "package", "resources", "validation", "help"}
excel_subs_characters_re = r"[:\*\?\/\[\]\\]"  # characters used in wildcards (MultiFile & MultiDataStore), forbidden in Excel sheet names
excel_subs_dest_character = '#'


def excel_name_of_sheet(resource_name: str) -> str:
        return re.sub(excel_subs_characters_re, excel_subs_dest_character, resource_name)

def excel_name_of_builder(resource_builder: BuilderResourceABC) -> str:
    if isinstance(resource_builder, BuilderMultiDataStore):
        return excel_name_of_sheet(resource_builder.name)
    else:
        return resource_builder.name


class BuilderPackageBasic:
    """
    Class to store an image of a CKAN package defined by an Excel worksheet


    __NB__: There are several paths to distinguish:

    - the path of the Excel worksheet
    - base_dir: the base directory for relative paths
    - resources_base_dir: the base directory for resources (for upload), which is generally defined relative to base_dir
    - out_dir: the output directory, for download, absolute or relative to the cwd (current working directory)

    __NB__: A builder can refer to the following external files:

    - CKAN API key file (.txt)
    - Proxy authentication file (.txt)
    - CKAN CA certificate file (.pem)
    - CA certificate for external connexions (.pem)
    - Data format policy file (.json)
    - External Python module (.py) containing DataFrame modification functions for upload/download of a DataStore
    """
    default_to_json_reduced_size:bool = False
    default_sample_url_suffix:str = "-sample"
    default_sample_title_suffix:str = " - Sample"

    def __init__(self, package_name:str=None, *, package_id:str=None,
                 title: str = None, description: str = None, private: bool = None, state: CkanState = None,
                 version: str = None,
                 url: str = None, tags: List[str] = None,
                 organization_name:str=None, license_name:str=None, src=None):
        if src is not None:
            src.copy(dest=self)
        if state is None:
            state = builder_default_package_state
        self.builder_source_file: Union[str, None] = None
        self.builder_format_version: Union[str, None] = None
        # package attributes
        self.package_attributes: CkanPackageInfo = CkanPackageInfo(package_name=package_name, package_id=package_id,
                                                                   title=title, description=description, private=private, state=state,
                                                                   version=version, url=url, tags=tags)
        self.package_attributes.state = state
        # if private is None:
        #     self.package_attributes.private = True  # package private by default
        self.organization_name: Union[str, None] = organization_name
        self.license_name: Union[str, None] = license_name
        # package resources
        self._resources_base_dir_src: Union[str, None] = None  # source of the resources_base_dir
        self._resources_base_dir: Union[str, None] = None
        self.resource_builders:OrderedDict[str,BuilderResourceABC] = OrderedDict()
        self._default_out_dir_src: Union[str, None] = None
        self._default_out_dir: Union[str, None] = None
        # auxiliary builders
        self.ckan_builder: BuilderCkan = BuilderCkan()
        self.external_python_code: Union[PythonUserCode, None] = None
        self.comment: str = ""
        self.built_from_ckan:bool = False

    def __str__(self):
        return f"Package builder for {self.package_name} ({len(self.resource_builders)} resources)"

    def copy(self, dest=None) -> "BuilderPackageBasic":
        if dest is None:
            dest = BuilderPackageBasic()
        dest.builder_source_file = self.builder_source_file
        dest.builder_format_version = self.builder_format_version
        dest.package_attributes = self.package_attributes.copy()
        dest.organization_name = self.organization_name
        dest.license_name = self.license_name
        dest._resources_base_dir_src = self._resources_base_dir_src
        dest._resources_base_dir = self._resources_base_dir
        dest._default_out_dir_src = self._default_out_dir_src
        dest._default_out_dir = self._default_out_dir
        dest.resource_builders = OrderedDict()
        dest.comment = self.comment
        dest.built_from_ckan = self.built_from_ckan
        for resource_name, resource_builder in self.resource_builders.items():
            dest.resource_builders[resource_name] = resource_builder.copy(parent=dest)  # change parent reference to the new instance of BuilderPackage
            assert_or_raise(dest.resource_builders[resource_name].parent_package == dest, RuntimeError())
        dest.ckan_builder = self.ckan_builder.copy()
        if self.external_python_code is not None:
            dest.external_python_code = self.external_python_code.copy()
        return dest

    def __copy__(self):
        return self.copy()

    def __del__(self):
        self.clear_secrets_and_disconnect()

    # Context Manager behavior ----------
    # to use PackageBuilder in a "with" statement
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.clear_secrets_and_disconnect()
        return True
    # -------------

    @staticmethod
    def setup_auto_draft_state(mode_auto:bool=None, *, draft_state_by_default:bool=None) -> None:
        """
        By default, packages are created in Draft state. This function disables this feature.
        Call before instantiating any package builder (BuilderPackage).

        :param mode_auto: set to True/False to setup at the same time the package state during upload
            and the default package state, applied at the end of the upload, if not specified by the user in the Excel workbook.
        :param draft_state_by_default: specific setting for the default package state (applied at the end of the upload).
        """
        global builder_default_package_state, initial_package_building_state
        if mode_auto is not None:
            if mode_auto:
                builder_default_package_state = CkanState.Draft
                initial_package_building_state = CkanState.Draft
            else:
                builder_default_package_state = CkanState.Active
                initial_package_building_state = None
        if draft_state_by_default is not None:
            builder_default_package_state = CkanState.Draft if draft_state_by_default else CkanState.Active

    def clear_secrets_and_disconnect(self) -> None:
        for resource_builder in self.resource_builders.values():
            resource_builder.clear_secrets_and_disconnect()

    def _check_mandatory_attributes(self):
        if self.package_name is None:
            raise MandatoryAttributeError("Package", "name")
        # organization can be non-mandatory depending on CKAN configuration
        # if self.organization_name is None:
        #     raise MissingMandatoryAttributeError("Package", "owner_org")

    def clear_ids(self):
        """
        Clear all known ids from package and resource builders
        :return:
        """
        self.package_attributes.id = None
        for resource_builder in self.resource_builders.values():
            resource_builder.known_id = None

    @staticmethod
    def unlock_external_code_execution(value:bool=True):
        """
        This function enables external code execution for the PythonUserCode class.
        It is necessary to load builders which specify an Auxiliary functions file.

        __Warning__:
        only run code if you trust the source!

        :return:
        """
        unlock_external_code_execution(value)

    @staticmethod
    def unlock_no_ca(value:bool=True):
        """
        This function enables you to disable the CA verification of the CKAN server.

        __Warning__:
        Only allow in a local environment!

        """
        unlock_no_ca(value)

    @staticmethod
    def unlock_external_url_resource_download(value:bool=True):
        """
        This function enables the download of resources external from the CKAN server.
        """
        unlock_external_url_resource_download(value)

    @property
    def package_name(self) -> str:
        return self.package_attributes.name
    @package_name.setter
    def package_name(self, value:str):
        self.package_attributes.name = value
        self.update_package_name_in_resources()

    @property
    def resources_base_dir(self) -> str:
        return self._resources_base_dir
    def set_resources_base_dir(self, value:str, base_dir:str=None):
        self._resources_base_dir_src = value
        self._apply_resources_base_dir_src(base_dir=self.get_base_dir(base_dir=base_dir))

    @property
    def default_out_dir(self) -> str:
        return self._default_out_dir
    def set_default_out_dir(self, value:str, base_dir:str=None):
        self._default_out_dir_src = value
        self._apply_out_dir_src(base_dir=self.get_base_dir(base_dir=base_dir))

    def update_package_name_in_resources(self):
        """
        Update package_name attribute in resource_builders
        Call before any operation on resources.
        This function is marked as deprecated. It double-checks the reciprocal link between the package and its resources.
        """
        package_name = self.package_name
        for resource_builder in self.resource_builders.values():
            resource_builder.package_name = package_name  # performs double check - will be removed in future versions

    def init_resources_options_and_metadata(self, ckan:CkanApi, *, base_dir:str=None) -> None:
        """
        Update ckan options in resource_builders
        Call before any operation on resources
        """
        for resource_builder in self.resource_builders.values():
            resource_builder.package_name = self.package_name  # performs double check - will be removed in future versions
            resource_builder.init_options_from_ckan(ckan, base_dir=base_dir)
            self._init_resource_from_df_aux_fun(resource_builder)

    def _apply_resources_base_dir_src(self, base_dir:str):
        """
        The resources base directory is specified in a field of the Excel workbook.
        This function resolves the directory name, based on the location of the Excel file
        or the base_dir, if provided.

        :param base_dir:
        :return:
        """
        resources_base_dir_src = self._resources_base_dir_src
        if resources_base_dir_src is None:
            resources_base_dir = base_dir
        else:
            resources_base_dir_src = os.path.expanduser(resources_base_dir_src)
            if os.path.isabs(resources_base_dir_src):
                resources_base_dir = resources_base_dir_src
            else:
                assert(base_dir is not None)
                self._resources_base_dir_src = os.path.join(base_dir, resources_base_dir_src)
                resources_base_dir = self._resources_base_dir_src
        if resources_base_dir is not None and not os.path.isdir(resources_base_dir):
            if not os.path.exists(resources_base_dir):
                raise FileOrDirNotExistError(resources_base_dir)
            # the field points to a text file containing the resources_base_dir
            with open(resources_base_dir, "r") as f:
                resources_base_dir = f.readline().strip()
                f.close()
        self._resources_base_dir = sanitize_path(resources_base_dir)

    def _get_resources_base_dir_src(self, base_dir:str):
        return make_path_relative(self._resources_base_dir, base_dir)
        # elif self._resources_base_dir_src is not None and os.path.exists(self._resources_base_dir_src) and not os.path.isdir(self._resources_base_dir_src):
        #     return self._resources_base_dir_src if base_dir is None else os.path.relpath(self._resources_base_dir_src, base_dir)
        # else:
        #     return self._resources_base_dir if base_dir is None else os.path.relpath(self._resources_base_dir, base_dir)

    def _apply_out_dir_src(self, base_dir:str, not_exist_error:bool=False):
        """
        The default download directory is specified in a field of the Excel workbook.
        This function resolves the directory name, based on the location of the Excel file
        or the base_dir, if provided.

        :param base_dir:
        :return:
        """
        out_dir_src = self._default_out_dir_src
        if out_dir_src is None:
            out_dir = None  # by default, do not define an output dir
        else:
            out_dir_keyword = out_dir_src.lower().strip()
            out_dir_src = os.path.expanduser(out_dir_src)
            if out_dir_keyword == "none":
                out_dir = None  # by default, do not define an output dir
            elif os.path.isabs(out_dir_src):
                out_dir = out_dir_src
            else:
                assert(base_dir is not None)
                self._default_out_dir_src = os.path.join(base_dir, out_dir_src)
                out_dir = self._default_out_dir_src
        if out_dir is not None and not os.path.isdir(out_dir):
            if not os.path.exists(out_dir):
                if not_exist_error:
                    raise FileOrDirNotExistError(out_dir)
                else:
                    msg = f"Default output directory {out_dir} does not exist! It will be created if you call the download function with no out_dir."
                    warn(msg)
                    self._default_out_dir = out_dir
                    return
            # the field points to a text file containing the out_dir
            with open(out_dir, "r") as f:
                out_dir = f.readline().strip()
                f.close()
        self._default_out_dir = sanitize_path(out_dir)

    def _get_out_dir_src(self, base_dir:str):
        return make_path_relative(self._default_out_dir_src, base_dir)


    def _load_from_df(self, info_df: pd.DataFrame, package_df: pd.DataFrame, base_dir:str=None) -> None:
        """
        Function to load builder parameters from a DataFrame, usually from an Excel worksheet

        :param package_df:
        :return:
        """
        if info_df is not None:
            package_df = pd.concat([package_df, info_df], axis=1)
        original_columns = list(package_df.columns)
        package_df.columns = package_df.columns.map(str.lower)
        package_df.columns = package_df.columns.map(str.strip)
        renamed_columns = list(package_df.columns)
        # info
        base_dir = self.get_base_dir(base_dir=base_dir)
        if "builder format version" in package_df.columns:
            self.builder_format_version = _string_from_element(package_df.pop("builder format version")).strip()
            assert_or_raise(self.builder_format_version == BUILDER_VER, UnsupportedBuilderVersionError(self.builder_format_version))
        if "resources local directory" in package_df.columns:
            resources_base_dir_src = sanitize_path(_string_from_element(package_df.pop("resources local directory")))
        else:
            resources_base_dir_src = None
        self._resources_base_dir_src = resources_base_dir_src
        self._apply_resources_base_dir_src(base_dir=base_dir)
        if "download directory" in package_df.columns:
            out_dir_src = sanitize_path(_string_from_element(package_df.pop("download directory")))
        else:
            out_dir_src = None
        self._default_out_dir_src = out_dir_src
        self._apply_out_dir_src(base_dir=base_dir)
        if "auxiliary functions file" in package_df.columns:
            auxiliary_functions_file = sanitize_path(_string_from_element(package_df.pop("auxiliary functions file")))
            if auxiliary_functions_file is not None:
                self.external_python_code = PythonUserCode(auxiliary_functions_file, base_dir=base_dir)
        if "comment" in package_df.columns:
            self.comment = _string_from_element(package_df.pop("comment"), empty_value="")
        # package attributes
        self.package_attributes: CkanPackageInfo
        if "name" in package_df.columns:
            # deprecated: package "Name" column was renamed "Name in URL"
            self.package_name = _string_from_element(package_df.pop("name")).strip()
            msg = "Package attribute 'Name' was renamed to 'Name in URL'. Previous name 'Name' is deprecated. Use 'Name in URL' instead."
            warn(msg)
        else:
            self.package_name = _string_from_element(package_df.pop("name in url")).strip()
        self.package_attributes.title = _string_from_element(package_df.pop("title"))
        if "known id" in package_df.columns:
            self.package_attributes.id = _string_from_element(package_df.pop("known id"))
        if "description" in package_df.columns:
            self.package_attributes.description = _string_from_element(package_df.pop("description"))
        if "version" in package_df.columns:
            self.package_attributes.version = _string_from_element(package_df.pop("version"))
        if "visibility" in package_df.columns:
            visibility = _string_from_element(package_df.pop("visibility"))
            if visibility is not None:
                self.package_attributes.private = CkanVisibility.from_str(visibility).to_bool_is_private()
        if "state" in package_df.columns:
            state = _string_from_element(package_df.pop("state"))
            if state is not None:
                self.package_attributes.state = CkanState.from_str(state)
        if "url" in package_df.columns:
            # field not in the default Excel file
            self.package_attributes.url = _string_from_element(package_df.pop("url"))
        if "tags" in package_df.columns:
            tags_string = _string_from_element(package_df.pop("tags"))
            if tags_string is not None:
                self.package_attributes.tags = [label.strip() for label in tags_string.split(ckan_tags_sep)]
        if "author" in package_df.columns:
            self.package_attributes.author = _string_from_element(package_df.pop("author"))
        if "author email" in package_df.columns:
            self.package_attributes.author_email = _string_from_element(package_df.pop("author email"))
        if "maintainer" in package_df.columns:
            self.package_attributes.maintainer = _string_from_element(package_df.pop("maintainer"))
        if "maintainer email" in package_df.columns:
            self.package_attributes.maintainer_email = _string_from_element(package_df.pop("maintainer email"))
        # fields which may require additional CKAN requests to obtain ids of the designated objects
        if "license" in package_df.columns:
            self.license_name = _string_from_element(package_df.pop("license"))
        if "organization" in package_df.columns:
            self.organization_name = _string_from_element(package_df.pop("organization"))
        # other fields = user custom fields
        if "attribute" in package_df.columns:
            package_df.pop("attribute")  # reserved name for table header
        remaining_columns = list(package_df.columns)
        for column in remaining_columns:
            original_column = original_columns[renamed_columns.index(column)]
            self.package_attributes.custom_fields[original_column] = _string_from_element(package_df[column])
        self._check_mandatory_attributes()

    def _to_dict(self, base_dir:str=None, include_id:bool=True) -> Tuple[dict, dict]:
        """
        Function to export builder parameters to an Excel worksheet, using the same fields as the input format

        :see: _load_from_df
        :see: to_xls
        :return:
        """
        info_dict = OrderedDict()
        info_dict["Builder format version"] = BUILDER_VER
        info_dict["Auxiliary functions file"] = make_path_relative(self.external_python_code.python_file, to_base_dir=base_dir) if self.external_python_code is not None else ""
        info_dict["Resources local directory"] = self._get_resources_base_dir_src(base_dir=base_dir)
        info_dict["Download directory"] = self._get_out_dir_src(base_dir=base_dir)
        info_dict["Comment"] = self.comment
        package_dict = OrderedDict()
        package_dict["Name in URL"] = self.package_name
        package_dict["Title"] = self.package_attributes.title
        if include_id and self.package_attributes.id:
            package_dict["Known Id"] = self.package_attributes.id
        package_dict["Description"] = self.package_attributes.description if self.package_attributes.description is not None else ""
        package_dict["Version"] = self.package_attributes.version if self.package_attributes.version is not None else ""
        package_dict["Visibility"] = CkanVisibility.from_bool_is_private(self.package_attributes.private).name if self.package_attributes.private is not None else ""
        package_dict["State"] = self.package_attributes.state.name if self.package_attributes.state is not None else ""
        package_dict["Organization"] = self.organization_name if self.organization_name is not None else ""
        package_dict["License"] = self.license_name if self.license_name is not None else ""
        package_dict["URL"] = self.package_attributes.url if self.package_attributes.url is not None else ""
        package_dict["Tags"] = ckan_tags_sep.join(self.package_attributes.tags) if self.package_attributes.tags is not None else ""
        package_dict["Author"] = self.package_attributes.author if self.package_attributes.author is not None else ""
        package_dict["Author Email"] = self.package_attributes.author_email if self.package_attributes.author_email is not None else ""
        package_dict["Maintainer"] = self.package_attributes.maintainer if self.package_attributes.maintainer is not None else ""
        package_dict["Maintainer Email"] = self.package_attributes.maintainer_email if self.package_attributes.maintainer_email is not None else ""
        for key, value in self.package_attributes.custom_fields.items():
            package_dict[key] = value if value is not None else ""
        return info_dict, package_dict

    def _get_builder_df_help_dict(self) -> Tuple[dict, dict]:
        info_help_dict = {
            "Builder format version": "Version of the file format for the script that processes this file",
            "Auxiliary functions file": "Path to a Python file containing auxiliary functions, relative to this Excel workbook folder\n"
                                        + "Warning: only execute code if you trust the source !",
            "Resources local directory": "Path to the local directory containing the resources to upload or text file defining this directory, relative to this Excel workbook folder",
            "Download directory": "Default path to download the resources to, relative to this Excel workbook folder",
            "Comment": "Place to add a comment on this file",
        }
        package_help_dict = {
            "Name in URL": "Name used in the URL to reference the dataset (short name, only accepting lowercase letters (a-z), digits (0-9), `-` and `_`, no accents, no spaces, length 2-100 characters)",
            "Title": "Title of the resource",
            "Description": "Description can use Markdown formatting",
            "Visibility": "Private/Public",
            "State": "Active/Draft/Deleted",
            "Organization": "Organization title, name or ID (mandatory)",
            "License": "License title or ID",
            "URL": "A URL for the dataset's source",
            "Tags": "Comma-separated list of tags (refer to data format policy)",
            "Author": "Recommended to specify author/maintainer to communicate on the dataset",
            "Maintainer": "Recommended to specify author/maintainer to communicate on the dataset",
        }
        if self.package_attributes.id:
            package_help_dict["Known Id"] = "ID of the resource in the CKAN database, last requested"
        package_help_dict.update({key: "Custom key-value pair (refer to data format policy)" for key in self.package_attributes.custom_fields.keys()})
        return info_help_dict, package_help_dict

    def _load_from_dict(self, info_dict: dict, package_dict: dict, base_dir:str=None) -> None:
        if info_dict is not None:
            info_df = pd.DataFrame([info_dict], index=["Value"])
            info_df = info_df.transpose()
            info_df.index.name = "Attribute"
            info_df = info_df.transpose()
        else:
            info_df = None
        package_df = pd.DataFrame([package_dict], index=["Value"])
        package_df = package_df.transpose()
        package_df.index.name = "Attribute"
        package_df = package_df.transpose()
        self._load_from_df(info_df, package_df, base_dir=base_dir)

    def _get_builder_df(self, base_dir:str=None, include_id:bool=True) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Converts the result of method _to_dict() into a DataFrame

        :return:
        """
        info_dict, package_dict = self._to_dict(base_dir=base_dir, include_id=include_id)
        info_help_dict, package_help_dict = self._get_builder_df_help_dict()
        package_df = pd.DataFrame([package_dict, package_help_dict], index=["Value", "Help"])
        package_df = package_df.transpose()
        package_df.index.name = "Attribute"
        info_df = pd.DataFrame([info_dict, info_help_dict], index=["Value", "Help"])
        info_df = info_df.transpose()
        info_df.index.name = "Attribute"
        return info_df, package_df

    def _check_resource_duplicates(self):
        duplicates = find_duplicates([resource_builder.name for resource_builder in self.resource_builders.values()])
        if len(duplicates) > 0:
            raise DuplicateNameError("Resource", duplicates)

    def _get_resources_dict(self, include_id:bool=True) -> Dict[str, dict]:
        self._check_resource_duplicates()
        resources_dict = {resource_builder.name: resource_builder._to_dict(include_id=include_id) for resource_builder in self.resource_builders.values()}
        return resources_dict

    def _get_resources_df(self, include_id:bool=True) -> pd.DataFrame:
        """
        Calls the method _to_dict() on all resources and returns the DataFrame listing the resources of the package

        :return:
        """
        resources_dict_list = [value for value in self._get_resources_dict(include_id=include_id).values()]
        resources_df = pd.DataFrame.from_records(resources_dict_list)
        return resources_df

    def _get_datastores_dict(self) -> Dict[str, dict]:
        """
        Calls the method _get_fields_dict() on all resources which are DataStores and returns a DataFrame per DataStore
        listing the fields of the DataStore with their metadata

        :return:
        """
        return {resource.name: resource._get_fields_dict() for resource in self.resource_builders.values()
                if (isinstance(resource, BuilderDataStoreABC) or isinstance(resource, BuilderMultiDataStore)) and resource.field_builders_user is not None}

    def _get_datastores_df(self) -> Dict[str, pd.DataFrame]:
        """
        Calls the method _get_fields_df() on all resources which are DataStores and returns a DataFrame per DataStore
        listing the fields of the DataStore with their metadata

        :return:
        """
        return {resource.name: resource._get_fields_df() for resource in self.resource_builders.values()
                if (isinstance(resource, BuilderDataStoreABC) or isinstance(resource, BuilderMultiDataStore)) and resource.field_builders_user is not None}

    def get_all_df(self, base_dir:str=None, include_id:bool=True) -> Dict[str, pd.DataFrame]:
        """
        Returns all the dataframes used to define the object and components

        :return:
        """
        info_df, package_df = self._get_builder_df(base_dir=base_dir, include_id=include_id)
        ckan_df = self.ckan_builder._get_builder_df(base_dir=base_dir)
        resources_df = self._get_resources_df(include_id=include_id)
        datastores_df = self._get_datastores_df()
        df_dict = {"info": info_df, "ckan": ckan_df, "package": package_df, "resources": resources_df}
        df_dict.update(datastores_df)
        return df_dict

    def to_excel(self, path_or_buffer, *, engine:str=None, include_id:bool=True, include_help:bool=True, **kwargs) -> None:
        """
        Call this function to export the builder parameters to an Excel worksheet

        :param path_or_buffer:
        :param engine:
        :return:
        """
        if isinstance(path_or_buffer, str):
            base_dir, _ = os.path.split(path_or_buffer)
        else:
            base_dir = None
        info_df, package_df = self._get_builder_df(base_dir=base_dir, include_id=include_id)
        ckan_df = self.ckan_builder._get_builder_df(base_dir=base_dir)
        resources_df = self._get_resources_df(include_id=include_id)
        datastores_df = self._get_datastores_df()
        with pd.ExcelWriter(path_or_buffer, engine=engine, **kwargs) as writer:
            ckan_df.to_excel(writer, sheet_name="ckan", index=True)
            info_df.to_excel(writer, sheet_name="info", index=True)
            package_df.to_excel(writer, sheet_name="package", index=True)
            resources_df.to_excel(writer, sheet_name="resources", index=False)
            for name, df in datastores_df.items():
                resource_builder = self.resource_builders[name]
                sheet_name = resource_builder.columns_sheet_name if resource_builder.columns_sheet_name is not None else excel_name_of_sheet(name)
                df.to_excel(writer, sheet_name=sheet_name, index=False)
            if include_help:
                help_df = load_help_page_df()
                help_df.to_excel(writer, sheet_name="help", index=False, header=False)
            # writer.close()

    def to_dict(self, base_dir:str=None, include_id:bool=True, separate_field_builders:bool=False) -> dict:
        """
        Call this function to export the builder parameters to an Excel worksheet

        :return:
        """
        d = OrderedDict()
        d["Info"], d["Package"] = self._to_dict(base_dir=base_dir, include_id=include_id)
        d["CKAN"] = self.ckan_builder._to_dict(base_dir=base_dir)
        resources_dict: dict = self._get_resources_dict(include_id=include_id)
        datastores_dict = self._get_datastores_dict()
        if separate_field_builders:
            for name, fields_dict in datastores_dict.items():
                resource_builder = self.resource_builders[name]
                sheet_name = resource_builder.columns_sheet_name if resource_builder.columns_sheet_name is not None else excel_name_of_sheet(name)
                resources_dict[sheet_name] = list(fields_dict.values())
        else:
            for name, fields_dict in datastores_dict.items():
                resources_dict[name]["fields"] = list(fields_dict.values())
        d["Resources"] = list(resources_dict.values())
        return d

    def to_json(self, json_file:str, *, include_id:bool=True, reduced_size:bool=None) -> None:
        if reduced_size is None:
            reduced_size = self.default_to_json_reduced_size
        base_dir, _ = os.path.split(json_file)
        builder_dict = self.to_dict(base_dir=base_dir, include_id=include_id)
        with open(json_file, "w", encoding="utf-8") as f:
            if reduced_size:
                json.dump(builder_dict, f, ensure_ascii=False)
            else:
                json.dump(builder_dict, f, ensure_ascii=False, indent=4)
            f.close()

    def to_jsons(self, *, base_dir:str=None, include_id:bool=True, reduced_size:bool=None) -> str:
        if reduced_size is None:
            reduced_size = self.default_to_json_reduced_size
        builder_dict = self.to_dict(base_dir=base_dir, include_id=include_id)
        if reduced_size:
            return json.dumps(builder_dict, ensure_ascii=False)
        else:
            return json.dumps(builder_dict, ensure_ascii=False, indent=4)

    @staticmethod
    def from_ckan(ckan: CkanApiMap, package_info: Union[CkanPackageInfo, str],
                  *, base_dir:str=None, error_duplicates:bool=True) -> "BuilderPackageBasic":
        """
        Function to initialize a BuilderPackageBasic from information requested by the CKAN API

        :param ckan:
        :param package_info: The package to import or the package name
        :return:
        """
        if isinstance(package_info, str):
            package_info = ckan.get_package_info_or_request(package_info, datastore_info=True)
        package_info: CkanPackageInfo
        mdl = BuilderPackageBasic()
        mdl.package_attributes = package_info
        mdl.organization_name = package_info.organization_info.get_owner_org() if package_info.organization_info is not None else None
        mdl.license_name = package_info.license_id if package_info.license_id else None
        mdl.license_name = mdl.get_license_name(ckan)
        for resource_info in package_info.package_resources.values():
            if resource_info.name in mdl.resource_builders.keys():
                msg = DuplicateNameError("resource_builder", [resource_info.name])
                if error_duplicates:
                    raise msg
                else:
                    warn(str(msg))
            mdl.resource_builders[resource_info.name] = init_resource_from_ckan(ckan, resource_info, parent=mdl)
        mdl.update_package_name_in_resources()
        mdl.init_resources_options_and_metadata(ckan, base_dir=base_dir)
        mdl.builder_source_file = "ckan"
        mdl.built_from_ckan = True
        return mdl

    def update_from_ckan(self, ckan:CkanApiMap, *, error_not_found:bool=True) -> None:
        """
        Update IDs from CKAN mapped objects.
        Objects must be mapped first.
        """
        package_info = ckan.map.get_package_info(self.package_name, error_not_mapped=error_not_found)
        package_id = package_info.id
        self.package_attributes.id = package_id
        for resource_builder in self.resource_builders.values():
            resource_info = ckan.map.get_resource_info(resource_builder.name, package_id, error_not_mapped=error_not_found)
            resource_builder.id = resource_info.id if resource_info is not None else None

    def _init_resource_from_df_aux_fun(self, resource_builder: BuilderResourceABC) -> None:
        if isinstance(resource_builder, BuilderDataStoreABC):
            resource_builder.df_mapper._connect_aux_functions(self.external_python_code,
                                                              aux_upload_fun_name=resource_builder.aux_upload_fun_name,
                                                              aux_download_fun_name=resource_builder.aux_download_fun_name)
            if isinstance(resource_builder.local_file_format, UserFileFormat):
                resource_builder.local_file_format._connect_aux_functions(self.external_python_code,
                                                                          aux_read_fun_name=resource_builder.aux_read_fun_name,
                                                                          aux_write_fun_name=resource_builder.aux_write_fun_name)

    def to_ckan_package_info(self, *, check_id:bool=True) -> CkanPackageInfo:
        """
        Function to insert the information coming from the builder into the CKAN map.
        Requires the IDs of the package and resources to be known.
        This enables to use the stored IDs instead of querying the CKAN API for these IDs.

        :return:
        """
        package_id = self.package_attributes.id
        package_info: CkanPackageInfo = self.package_attributes.copy()
        if package_id is None and check_id:
            msg = MissingIdError("package", self.package_name)
            raise(msg)
        for resource_builder in self.resource_builders.values():
            if isinstance(resource_builder, BuilderMultiFile):
                msg = f"Multi-resource builder is not compatible with updating CKAN resource ids from known ids because more than one id is expected (resource builder {resource_builder.name})"
                warn(msg)
            else:
                package_info.package_resources[resource_builder.known_id] = resource_builder._to_ckan_resource_info(package_id, check_id=check_id)
        package_info.resources_id_index = {resource_info.name: resource_info.id for resource_info in package_info.package_resources.values()}  # resource name -> id
        package_info.resources_id_index_counts = {}  # resource name -> counter
        for resource_info in package_info.package_resources.values():
            if resource_info.name not in package_info.resources_id_index_counts.keys():
                package_info.resources_id_index_counts[resource_info.name] = 1
            else:
                package_info.resources_id_index_counts[resource_info.name] += 1
        return package_info

    def update_ckan_map(self, ckan: CkanApiMap, *, warn_msg:bool=True) -> CkanPackageInfo:
        """
        This function updates the CKAN map from the information contained in this builder.
        For this to work, the package and resource ids must be known.
        This is not the case if the package was not initialized.
        Use if the builder was initialized from ckan or use with precaution.

        .. warning :: This function bypasses the ids which should normally be obtained through the API.
            Use at your own risk.

        :param ckan:
        :return:
        """
        if warn_msg:
            msg = "update_ckan_map overrides the ids which should normally be obtained through the API. Use at your own risk."
            warn(msg)
        package_info = self.to_ckan_package_info(check_id=True)
        ckan.map._update_package_info(package_info)
        return package_info.copy()

    def map_resources(self, ckan: CkanApiMap, *, error_not_found:bool=True, cancel_if_exists:bool=True,
                      datastore_info:bool=True) -> Union[CkanPackageInfo,None]:
        """
        proxy call to ckan.map_resources and returns package information from CKAN

        :param ckan:
        :param error_not_found:
        :param cancel_if_exists:
        :return:
        """
        ckan.map_resources(self.package_name, datastore_info=datastore_info, error_not_found=error_not_found, only_missing=cancel_if_exists)
        package_info = ckan.map.get_package_info(self.package_name, error_not_mapped=error_not_found)
        if package_info is None:
            return None
        self.package_attributes.id = package_info.id
        self.update_from_ckan(ckan, error_not_found=error_not_found)
        return package_info

    def _load_package_resources_list_df(self, resources_df: pd.DataFrame, base_dir:str=None) -> None:
        resources_df.columns = resources_df.columns.map(str.lower)
        resources_df.columns = resources_df.columns.map(str.strip)
        self.resource_builders = OrderedDict()
        for row_loc, row in resources_df.iterrows():
            resource_builder = init_resource_from_df(row, base_dir=base_dir, parent=self)
            if resource_builder is not None:
                self._init_resource_from_df_aux_fun(resource_builder)
                if resource_builder.name in self.resource_builders.keys():
                    raise DuplicateNameError("resource_builder", [resource_builder.name])
                if resource_builder.name.lower() in forbidden_resource_names:
                    raise ForbiddenNameError("resource_builder", [resource_builder.name])
                self.resource_builders[resource_builder.name] = resource_builder
        # self._update_package_name_resources()  # call after full init in caller function

    @staticmethod
    def from_excel(path_or_stream, *, proxies:dict=None, engine:str=None, **kwargs) -> "BuilderPackageBasic":
        """
        Load package definition from an Excel workbook.

        :param path_or_stream: path to the Excel workbook
        :param engine: Engine used by pandas.read_excel(). Supported engines: xlrd, openpyxl, odf, pyxlsb, calamine.
        openpyxl makes part of this package's optional requirements
        :return:
        """
        warnings.filterwarnings('ignore', category=UserWarning, module='openpyxl')  # openpyxl produces warnings concerning data validation
        mdl = BuilderPackageBasic()
        mdl.builder_source_file = path_or_stream
        with open(path_or_stream, "rb") as stream:  # enforce read-only mode
            with pd.ExcelFile(stream, engine=engine, **kwargs) as xls:
                sheet_names = set(xls.sheet_names)
                sheet_names_lower_index = {sheet_name.lower().strip(): sheet_name for sheet_name in sheet_names}
                package_df = pd.read_excel(xls, sheet_name=sheet_names_lower_index["package"], header=None)
                package_df.set_index(0, inplace=True)
                assert(package_df.index.is_unique)  # verify integrity
                package_df = package_df.T
                if "info" in sheet_names_lower_index.keys():
                    info_df = pd.read_excel(xls, sheet_name=sheet_names_lower_index["info"], header=None)
                    info_df.set_index(0, inplace=True)
                    assert(info_df.index.is_unique)  # verify integrity
                    info_df = info_df.T
                else:
                    info_df = None
                base_dir = mdl.get_base_dir(None)
                mdl._load_from_df(info_df, package_df, base_dir=base_dir)
                if "ckan" in sheet_names_lower_index.keys():
                    ckan_df = pd.read_excel(xls, sheet_name=sheet_names_lower_index["ckan"], header=None)
                    ckan_df.set_index(0, inplace=True)
                    assert(ckan_df.index.is_unique)  # verify integrity
                    ckan_df = ckan_df.T
                    mdl.ckan_builder._load_from_df(ckan_df, base_dir=base_dir, proxies=proxies)
                resources_df = pd.read_excel(xls, sheet_name=sheet_names_lower_index["resources"])
                mdl._load_package_resources_list_df(resources_df, base_dir=base_dir)
                resource_sheets = sheet_names - {sheet_names_lower_index[name] for name in forbidden_resource_names if name in sheet_names_lower_index.keys()}
                used_resource_sheets = set()
                for resource_builder in mdl.resource_builders.values():
                    resource_sheet = None
                    equiv_name = excel_name_of_builder(resource_builder)
                    if resource_builder.columns_sheet_name is not None:
                        resource_sheet = resource_builder.columns_sheet_name.strip()
                        assert_or_raise(resource_sheet in resource_sheets, MissingDataStoreColumnsSheet(resource_builder.name, resource_sheet))
                    elif resource_builder.name in resource_sheets:
                        resource_sheet = resource_builder.name
                    elif equiv_name in resource_sheets:
                        resource_sheet = equiv_name
                    if resource_sheet is not None:
                        fields_df = pd.read_excel(xls, sheet_name=resource_sheet)
                        assert(isinstance(resource_builder, BuilderDataStoreABC) or isinstance(resource_builder, BuilderMultiDataStore))
                        resource_builder._load_fields_df(fields_df)
                        used_resource_sheets.add(resource_sheet)
                        # resource_builder.columns_sheet_name = resource_sheet
                mdl.update_package_name_in_resources()
                unused_resource_sheets = resource_sheets - used_resource_sheets
                if len(unused_resource_sheets) > 0:
                    msg = f"Sheets present but not used: {', '.join(unused_resource_sheets)}"
                    warn(msg)
                xls.close()
        return mdl

    @staticmethod
    def from_dict(d:dict, base_dir:str=None, *, proxies:dict=None) -> "BuilderPackageBasic":
        """
        Load package definition from a dictionary.
        In this case, the base directory used to specify the resources locations must be given manually.
        This is usually the directory of the file where the dictionary comes from.

        :param d:
        :param base_dir:
        :param proxies:
        :return:
        """
        mdl = BuilderPackageBasic()
        mdl.builder_source_file = None
        sheet_names = set(d.keys())
        sheet_names_lower_index = {sheet_name.lower().strip(): sheet_name for sheet_name in sheet_names}
        info_dict = d[sheet_names_lower_index["info"]] if "info" in sheet_names_lower_index.keys() else None
        mdl._load_from_dict(info_dict, d[sheet_names_lower_index["package"]], base_dir=base_dir)
        if "ckan" in sheet_names_lower_index.keys():
            ckan_dict = d[sheet_names_lower_index["ckan"]]
            mdl.ckan_builder._load_from_dict(ckan_dict, base_dir=base_dir, proxies=proxies)
        resources_dict = dict()
        for resource_dict in d[sheet_names_lower_index["resources"]]:
            resource_dict_alt = {k.lower().strip(): v for k, v in resource_dict.items()}
            resources_dict[resource_dict_alt["name"]] = resource_dict_alt
        resources_df = pd.DataFrame(list(resources_dict.values()))
        mdl._load_package_resources_list_df(resources_df, base_dir=base_dir)
        resource_sheets = sheet_names - {sheet_names_lower_index[name] for name in forbidden_resource_names if name in sheet_names_lower_index.keys()}
        used_resource_sheets = set()
        for resource_builder in mdl.resource_builders.values():
            if "fields" in resources_dict[resource_builder.name]:
                assert(isinstance(resource_builder, BuilderDataStoreABC) or isinstance(resource_builder, BuilderMultiDataStore))
                fields_df = pd.DataFrame(resources_dict[resource_builder.name]["fields"])
                resource_builder._load_fields_df(fields_df)
            else:
                resource_sheet = None
                equiv_name = excel_name_of_builder(resource_builder)
                if resource_builder.columns_sheet_name is not None:
                    resource_sheet = resource_builder.columns_sheet_name.strip()
                    assert_or_raise(resource_sheet in resource_sheets, MissingDataStoreColumnsSheet(resource_builder.name, resource_sheet))
                elif resource_builder.name in resource_sheets:
                    resource_sheet = resource_builder.name
                elif equiv_name in resource_sheets:
                    resource_sheet = equiv_name
                if resource_sheet is not None:
                    assert(isinstance(resource_builder, BuilderDataStoreABC) or isinstance(resource_builder, BuilderMultiDataStore))
                    fields_df = pd.DataFrame(list(d[resource_sheet].values()))
                    resource_builder._load_fields_df(fields_df)
                    used_resource_sheets.add(resource_sheet)
                    # resource_builder.columns_sheet_name = resource_sheet
        mdl.update_package_name_in_resources()
        unused_resource_sheets = resource_sheets - used_resource_sheets
        if len(unused_resource_sheets) > 0:
            msg = f"Sheets present but not used: {', '.join(unused_resource_sheets)}"
            warn(msg)
        return mdl

    @staticmethod
    def from_json(json_file, *, proxies:dict=None) -> "BuilderPackageBasic":
        base_dir, _ = os.path.split(json_file)
        with open(json_file, "r") as f:
            builder_dict = json.load(f)
            f.close()
        mdl = BuilderPackageBasic.from_dict(builder_dict, base_dir=base_dir, proxies=proxies)
        mdl.builder_source_file = json_file
        return mdl

    @staticmethod
    def from_jsons(stream:str, *, source_file:str=None, proxies:dict=None) -> "BuilderPackageBasic":
        base_dir, _ = os.path.split(source_file) if source_file is not None else (None, None)
        builder_dict = json.loads(stream)
        mdl = BuilderPackageBasic.from_dict(builder_dict, base_dir=base_dir, proxies=proxies)
        mdl.builder_source_file = source_file
        return mdl

    def get_owner_org(self, ckan: CkanApiMap) -> str:
        """
        Returns the owner organization for the package.
        The owner organization can be specified by its name, title or id

        :param ckan:
        :return:
        """
        if self.organization_name is not None:
            ckan.organization_list_all(cancel_if_present=True)
            # organization_info = ckan.get_organization_info_or_request(self.organization_name, error_not_found=True)
            organization_info = ckan.map.get_organization_info(self.organization_name, error_not_mapped=True)
            owner_org = organization_info.get_owner_org()
        else:
            owner_org = None
        return owner_org

    def get_license_id(self, ckan: CkanApiMap) -> str:
        """
        Returns the license for the package.
        The license can be specified by its title or id

        :param ckan:
        :return:
        """
        if self.license_name is not None:
            ckan.license_list(cancel_if_present=True)
            license_id = ckan.map.get_license_id(self.license_name, error_not_mapped=True)
        else:
            license_id = None
        return license_id

    def get_license_info(self, ckan: CkanApiMap) -> CkanLicenseInfo:
        license_id = self.get_license_id(ckan)
        license_info = ckan.map.get_license_info(license_id) if license_id is not None else None
        return license_info

    def get_license_name(self, ckan: CkanApiMap) -> str:
        license_info = self.get_license_info(ckan)
        return license_info.title if license_info is not None else None

    def get_package_page_url(self, ckan:CkanApi, *, error_not_found:bool=True, default_url:bool=False) -> str:
        return ckan.get_package_page_url(self.package_name, error_not_found=error_not_found, default_url=default_url)

    def patch_request_package(self, ckan:CkanApi) -> CkanPackageInfo:
        """
        Function to perform all the necessary requests to initiate/reupload the package on the CKAN server.
        This function does not upload the package resources.

        .. note :: The organization must be provided, especially if the package is private

        :param ckan:
        :return:
        """
        owner_org = self.get_owner_org(ckan)
        license_id = self.get_license_id(ckan)
        package_info = ckan.package_create(self.package_name, private=self.package_attributes.private,
            state=initial_package_building_state if initial_package_building_state is not None else self.package_attributes.state,
            title=self.package_attributes.title, notes=self.package_attributes.description, owner_org=owner_org,
            tags=self.package_attributes.tags, custom_fields=self.package_attributes.custom_fields,
            url=self.package_attributes.url, version=self.package_attributes.version,
            author=self.package_attributes.author, author_email=self.package_attributes.author_email,
            maintainer=self.package_attributes.maintainer, maintainer_email=self.package_attributes.maintainer_email,
            license_id=license_id,
            cancel_if_exists=True, update_if_exists=True, clear_if_deleted_state=None)
        if ckan.params.verbose_extra:
            print("Package created. URL:", self.get_package_page_url(ckan))
        return package_info

    def package_delete_resources(self, ckan: CkanApi, *, bypass_admin:bool=False):
        package_id = self.package_attributes.id
        ckan.package_delete_resources(package_id, bypass_admin=bypass_admin)

    def patch_request_full(self, ckan:CkanApi, *,
                           reupload:bool=False, override_ckan:bool=False, resources_base_dir:str=None,
                           create_default_view:bool=True, delete_all_resources:bool=False,
                           progress_callback:Union[CkanProgressCallbackABC,Callable]=None,
                           sample_df_dict:Dict[str, Union[bytes, pd.DataFrame]]=None) \
            -> Tuple[CkanPackageInfo, Dict[str, CkanResourceInfo]]:
        """
        Perform necessary requests to initiate/reupload the package and resources on the CKAN server.
        For folder resources, this only uploads the first file of the resource.

        :param ckan:
        :return:
        """
        # call to function update_request of package and update_request of resources
        if not isinstance(progress_callback, CkanProgressCallbackABC):
            progress_callback = CkanProgressCallback(progress_callback)
        if ckan.params.policy_check_pre:
            self.local_policy_check()
        resources_base_dir = self.get_resources_base_dir(resources_base_dir)
        self.upload_file_checks(resources_base_dir=resources_base_dir, ckan=ckan, verbose=True, raise_error=True)
        pkg_info = self.patch_request_package(ckan)
        ckan.map_resources(self.package_name, datastore_info=True)
        package_id = pkg_info.id
        self.package_attributes.id = package_id
        if delete_all_resources:
            ckan.package_delete_resources(package_id, bypass_admin=True)
        resource_info_dict: Dict[str, CkanResourceInfo] = {}
        self.update_package_name_in_resources()
        self.init_resources_options_and_metadata(ckan, base_dir=self.get_base_dir())
        if progress_callback is not None:
            progress_callback.start_task(len(self.resource_builders), level=CkanCallbackLevel.Resources)
        for resource_index, resource_builder in enumerate(self.resource_builders.values()):
            if progress_callback is not None:
                resource_builder.progress_callback = progress_callback
                progress_callback.task_progress(resource_index+1, len(self.resource_builders), level=CkanCallbackLevel.Resources)
            if create_default_view is not None:
                resource_builder.create_default_view = create_default_view
            if sample_df_dict is not None and resource_builder.name in sample_df_dict.keys():
                data_upload = sample_df_dict[resource_builder.name]
            else:
                data_upload = None
            if data_upload is not None:
                if isinstance(resource_builder, BuilderDataStoreABC):
                    assert(isinstance(data_upload, pd.DataFrame) or isinstance(data_upload, list))
                    resource_info = resource_builder.patch_request(ckan, package_id, reupload=reupload, override_ckan=override_ckan,
                                                                   resources_base_dir=resources_base_dir, df_upload=data_upload)
                else:
                    assert(isinstance(data_upload, bytes))
                    resource_info = resource_builder.patch_request(ckan, package_id, reupload=reupload, override_ckan=override_ckan,
                                                                   resources_base_dir=resources_base_dir, payload=data_upload)
            else:
                resource_info = resource_builder.patch_request(ckan, package_id, reupload=reupload, override_ckan=override_ckan,
                                                               resources_base_dir=resources_base_dir)
            resource_info_dict[resource_builder.name] = resource_info
            if resource_info is not None:  # this would be the case for BuilderMultiFile
                pkg_info.update_resource(resource_info)
            # else:
            #     assert(isinstance(resource_builder, BuilderMultiFile))
        self.package_resource_reorder(ckan)
        if ckan.params.policy_check_post:
            self.remote_policy_check(ckan)
        if progress_callback is not None:
            progress_callback.end_task(len(self.resource_builders), level=CkanCallbackLevel.Resources)
        return pkg_info, resource_info_dict

    def _get_mono_resource_used_files(self, resources_base_dir:str, ckan:CkanApi):
        """
        List files used by mono-resource builders

        :param resources_base_dir:
        :return:
        """
        mono_resource_used_files = set()
        for resource_builder in self.resource_builders.values():
            if isinstance(resource_builder, BuilderDataStoreMultiABC):
                if not isinstance(resource_builder, BuilderDataStoreHarvester):
                    file_list = resource_builder.init_local_files_list(resources_base_dir=resources_base_dir, ckan=ckan)
                    mono_resource_used_files.update(set(file_list))
            elif not (isinstance(resource_builder, BuilderMultiFile)):
                if resource_builder.get_sample_file_path(resources_base_dir, ckan) is not None and not resource_builder.sample_file_path_is_url():
                    mono_resource_used_files.add(resource_builder.get_sample_file_path(resources_base_dir, ckan))
        return mono_resource_used_files

    def upload_file_checks(self, resource_name:Union[str, List[str]]=None, *, resources_base_dir:str=None,
                           messages:Dict[str, ContextErrorLevelMessage]=None,
                           verbose:bool=True, raise_error:bool=False, ckan:CkanApi=None, **kwargs) -> bool:
        """
        Method to check the presence of all needed files before uploading or patching resources.

        :param resources_base_dir:
        :param ckan: Optional CkanApi object used to parameterize the requests to test the presence of resources defined by an url.
        :param kwargs: keyword arguments to specify connexion parameters for querying the urls.
        :return:
        """
        if resource_name is None:
            resource_name = list(self.resource_builders.keys())
        elif isinstance(resource_name, str):
            resource_name = [resource_name]
        if messages is None:
            messages = {}
        self.update_package_name_in_resources()
        resources_base_dir = self.get_resources_base_dir(resources_base_dir)
        mono_resource_used_files = self._get_mono_resource_used_files(resources_base_dir, ckan)
        for resource_builder_name in resource_name:
            resource_builder = self.resource_builders[resource_builder_name]
            if isinstance(resource_builder, BuilderMultiFile):
                messages[resource_builder_name] = resource_builder.upload_file_checks(resources_base_dir=resources_base_dir, ckan=ckan,
                                                        excluded_files=mono_resource_used_files if multi_file_exclude_other_files else None, **kwargs)
            else:
                messages[resource_builder_name] = resource_builder.upload_file_checks(resources_base_dir=resources_base_dir, ckan=ckan, **kwargs)
        num_messages = len([1 for message in messages.values() if message is not None])
        success = len([1 for message in messages.values() if message is not None and message.error_level == ErrorLevel.Error]) == 0
        if verbose and num_messages > 0:
            print("\n".join([f"for resource {key}: {message}" for key, message in messages.items() if message is not None]))
        if raise_error and not success:
            raise FileNotFoundError("\n".join([f"for resource {key}: {message}" for key, message in messages.items() if message is not None and message.error_level == ErrorLevel.Error]))
        return success

    def upload_large_datasets(self, ckan:CkanApi, *, resources_base_dir:str=None, threads:int=None,
                              progress_callback:Union[CkanProgressCallbackABC,Callable]=None,
                              only_missing:bool=False, from_line_count:bool=False, allow_chunks:bool=True) -> None:
        """
        Method to upload large datasets of the package.
        The small datasets are to be uploaded with the patch_request_full method.

        :param ckan:
        :param resources_base_dir:
        :param threads:
        :param progress_callback:
        :param only_missing: upsert only missing rows for DataStores and only missing files for MultiFile
        :param from_line_count: count the lines on the CKAN DataStore and ignore the first n lines of your data source
        :param allow_chunks: read DataStore files by chunks, when available
        :return:
        """
        if not isinstance(progress_callback, CkanProgressCallbackABC):
            progress_callback = CkanProgressCallback(progress_callback)
        if threads is None:
            threads = self.ckan_builder.default_thread_count
        self.info_request_package(ckan=ckan)
        resources_base_dir = self.get_resources_base_dir(resources_base_dir)
        self.update_package_name_in_resources()
        self.init_resources_options_and_metadata(ckan, base_dir=self.get_base_dir())
        resource_names = [key for key, resource_builder in self.resource_builders.items() if isinstance(resource_builder, BuilderDataStoreMultiABC)]
        self.upload_file_checks(resource_names, resources_base_dir=resources_base_dir, ckan=ckan, verbose=True, raise_error=True)
        mono_resource_used_files = self._get_mono_resource_used_files(resources_base_dir, ckan)
        if progress_callback is not None:
            progress_callback.start_task(len(self.resource_builders), level=CkanCallbackLevel.Resources)
        for resource_index, resource_builder in enumerate(self.resource_builders.values()):
            if isinstance(resource_builder, BuilderDataStoreMultiABC):
                if progress_callback is not None:
                    resource_builder.progress_callback = progress_callback
                    progress_callback.task_progress(resource_index, len(self.resource_builders), level=CkanCallbackLevel.Resources)
                resource_builder.upload_request_full(ckan=ckan, resources_base_dir=resources_base_dir, threads=threads,
                                                     allow_chunks=allow_chunks,
                                                     only_missing=only_missing, from_line_count=from_line_count)
        for index, resource_builder in enumerate(self.resource_builders.values()):
            if isinstance(resource_builder, BuilderMultiFile):
                if progress_callback is not None:
                    resource_builder.progress_callback = progress_callback
                    # progress_callback.call(index, len(self.resource_builders), level=CkanCallbackLevel.Resources)
                resource_builder.upload_request_full(ckan=ckan, resources_base_dir=resources_base_dir, threads=threads,
                                                     only_missing=only_missing, from_line_count=from_line_count,
                                                     allow_chunks=allow_chunks,
                                                     excluded_files=mono_resource_used_files if multi_file_exclude_other_files else None)
        self.package_resource_reorder(ckan)
        self.patch_request_final(ckan)
        if progress_callback is not None:
            progress_callback.end_task(len(self.resource_builders), level=CkanCallbackLevel.Resources)

    def patch_request_final(self, ckan:CkanApi):
        if initial_package_building_state is not None:
            ckan.package_patch(self.package_name, state=self.package_attributes.state)

    def download_resource_df(self, ckan:CkanApi, resource_name:str, search_all:bool=False, **kwargs) -> pd.DataFrame:
        """
        Proxy for download_sample_df for a DataStore
        """
        self.update_package_name_in_resources()
        self.init_resources_options_and_metadata(ckan, base_dir=self.get_base_dir())
        assert(isinstance(self.resource_builders[resource_name], BuilderDataStoreABC))
        return self.resource_builders[resource_name].download_sample_df(ckan=ckan, search_all=search_all, **kwargs)

    def download_resource(self, ckan:CkanApi, resource_name:str, full_download:bool=False, **kwargs) -> bytes:
        """
        Proxy for download_sample for a resource
        """
        self.update_package_name_in_resources()
        self.init_resources_options_and_metadata(ckan, base_dir=self.get_base_dir())
        return self.resource_builders[resource_name].download_sample(ckan=ckan, full_download=full_download, **kwargs)

    def get_or_query_resource_id(self, ckan:CkanApi, resource_name:str, error_not_found:bool=True) -> str:
        self.update_package_name_in_resources()
        self.init_resources_options_and_metadata(ckan, base_dir=self.get_base_dir())
        return self.resource_builders[resource_name].get_or_query_resource_id(ckan, error_not_found=error_not_found)

    def _get_mono_resource_names(self):
        """
        List resource names of mono-resource builders.

        :return:
        """
        return {resource_name for resource_name, resource_builder in self.resource_builders.items() if not isinstance(resource_builder, BuilderMultiFile)}

    def download_request_full(self, ckan:CkanApi, out_dir:str=None, enforce_none_out_dir:bool=False, resource_name:str=None, full_download:bool=False,
                              threads:int=None, skip_existing:bool=True, progress_callback:Callable=None,
                              force:bool=False, rm_dir:bool=False) -> None:
        """
        Downloads the full package resources into out_dir.

        :param ckan:
        :param out_dir: download directory
        :param rm_dir: remove directory if exists before downloading
        :param skip_existing: skip download of existing resources
        :param enforce_none_out_dir: if no out_dir is provided, True: files will not be saved after download, False: default output dir will be used, if defined
        :param resource_name:
        :param full_download: option to fully download the resources. If False, only a partial download is made.
        :param threads:
        :param progress_callback:
        :param force: option to bypass the enable_download attribute of resources
        :return:
        """
        if threads is None:
            threads = self.ckan_builder.default_thread_count
        out_dir = self.get_default_out_dir(out_dir, enforce_none=enforce_none_out_dir)
        if out_dir is not None and os.path.isdir(out_dir):
            if rm_dir:
                shutil.rmtree(out_dir)
        self.info_request_package(ckan=ckan)
        if resource_name is None:
            resource_builders = self.resource_builders
        else:
            resource_builders = {resource_name: self.resource_builders[resource_name]}
        self.update_package_name_in_resources()
        self.init_resources_options_and_metadata(ckan, base_dir=self.get_base_dir())
        mono_resource_names = self._get_mono_resource_names()
        for resource_builder in resource_builders.values():
            if skip_existing is not None:
                resource_builder.download_skip_existing = skip_existing
            if not (isinstance(resource_builder, BuilderDataStoreMultiABC) or isinstance(resource_builder, BuilderMultiFile)):
                resource_builder.download_request(ckan, out_dir=out_dir, full_download=full_download,
                                                  threads=threads, force=force)
        for resource_builder in resource_builders.values():
            if isinstance(resource_builder, BuilderDataStoreMultiABC):
                if progress_callback is not None:
                    resource_builder.progress_callback = progress_callback
                resource_builder.download_request(ckan, out_dir=out_dir, full_download=full_download,
                                                  threads=threads, force=force)
        for resource_builder in resource_builders.values():
            if isinstance(resource_builder, BuilderMultiFile):
                if progress_callback is not None:
                    resource_builder.progress_callback = progress_callback
                resource_builder.download_request(ckan, out_dir=out_dir, full_download=full_download,
                                                  threads=threads, force=force, excluded_resource_names=mono_resource_names)

    def download_sample_df(self, ckan:CkanApi, resource_name:str=None, *, search_all:bool=False,
                           **kwargs) -> Dict[str, pd.DataFrame]:
        """
        Download a sample DataFrame for the DataStore type resources.

        :param ckan:
        :param resource_name:
        :return:
        """
        self.info_request_package(ckan=ckan)
        if resource_name is None:
            resource_builders = self.resource_builders
        else:
            resource_builders = {resource_name: self.resource_builders[resource_name]}
        self.update_package_name_in_resources()
        self.init_resources_options_and_metadata(ckan, base_dir=self.get_base_dir())
        sample_df_dict = {}
        for resource_builder in resource_builders.values():
            if isinstance(resource_builder, BuilderDataStoreABC):
                sample_df_dict[resource_builder.name] = resource_builder.download_sample_df(ckan, search_all=search_all, **kwargs)
        return sample_df_dict

    def download_sample(self, ckan:CkanApi, resource_name:str=None, *,
                        datastores_as_df:bool=True, download_url_resources:bool=False,
                        include_files:bool=True, empty_files:bool=False,
                        search_all:bool=False,
                        **kwargs) -> Dict[str, Union[bytes, pd.DataFrame]]:
        """
        Download samples for all resources. Resources which are not DataStores are downloaded entirely as bytes.

        :param ckan:
        :param resource_name: option to restrict to a single resource
        :param datastores_as_df: Download DataStores as DataFrames (do not convert to bytes)
        :param download_url_resources: Option to download resources aiming for an external URL.
        :param include_files: Option to include resources which are files as bytes.
        :param empty_files: Option to force file contents to an empty file.
        :param search_all: Option to search all resources before downloading (only applies to DataStores).
        :param kwargs: applies to download_sample_df
        :return: a dictionary with a sample for each resource
        """
        self.info_request_package(ckan=ckan)
        if resource_name is None:
            resource_builders = self.resource_builders
        else:
            resource_builders = {resource_name: self.resource_builders[resource_name]}
        self.update_package_name_in_resources()
        self.init_resources_options_and_metadata(ckan, base_dir=self.get_base_dir())
        sample_df_dict = {}
        for resource_builder in resource_builders.values():
            if (not download_url_resources) and (isinstance(resource_builder, BuilderUrl)
                    or isinstance(resource_builder, BuilderDataStoreUrl)):
                sample_df_dict[resource_builder.name] = None
            elif isinstance(resource_builder, BuilderDataStoreABC) and datastores_as_df:
                sample_df_dict[resource_builder.name] = resource_builder.download_sample_df(ckan, search_all=search_all, **kwargs)
            elif include_files and not empty_files:
                sample_df_dict[resource_builder.name] = resource_builder.download_resource_bytes(ckan, search_all=search_all, **kwargs)
            elif empty_files:
                sample_df_dict[resource_builder.name] = bytes()
            else:
                sample_df_dict[resource_builder.name] = None
        return sample_df_dict

    def setup_sample_package(self, ckan: CkanApi, package_name:str=None, *,
                             sample_url_suffix:str=None, sample_title_suffix:str=None,
                             sample_df_dict: Dict[str, Union[bytes, pd.DataFrame]]=None, return_sample:bool=False, **kwargs) \
        -> Union["BuilderPackageBasic", Tuple["BuilderPackageBasic", Dict[str, Union[bytes, pd.DataFrame]]]]:
        """
        Returns a package builder configured to represent a sample of the current package builder.
        Limitation: the current package builder must be created from CKAN.

        :param ckan:
        :param package_name: If specified, derives the package metadata from the specified package name. By default, the current package builder will be used.
        :param sample_url_suffix: Suffix to add to the package_name (default is "-sample")
        :param sample_title_suffix: Suffix to add to the package title (default is " - Sample")
        :param sample_df_dict: Option to transmit the data of each resource to the output of function.
        :param return_sample: Option to return the data of each resource.
        :param kwargs: Optional arguments to pass to the download_sample function.
        :return: a package builder configured to represent a sample of the current package builder.
            Optionally, the dictionary of resources to transmit
        """
        if sample_url_suffix is None:
            sample_url_suffix = BuilderPackageBasic.default_sample_url_suffix
        if sample_title_suffix is None:
            sample_title_suffix = BuilderPackageBasic.default_sample_title_suffix
        if package_name is None:
            mdl = self
        else:
            mdl = BuilderPackageBasic.from_ckan(ckan, package_name)
        if not mdl.built_from_ckan:
            msg = "To setup a sample package, the source package must be built from CKAN (for implementation reasons)"
            raise RuntimeError(msg)
        sample_mdl = mdl.copy()
        sample_mdl.package_name = mdl.package_name + sample_url_suffix
        sample_mdl.package_attributes.title = mdl.package_attributes.title + sample_title_suffix
        sample_mdl.package_attributes.url = ckan.get_package_page_url(mdl.get_or_query_package_id(ckan))  # mark as source
        sample_mdl.package_attributes.state = CkanState.Active  # publish at the end of the process
        for resource_builder in sample_mdl.resource_builders.values():
            resource_builder.aliases = []
        sample_mdl.clear_ids()
        if return_sample and sample_df_dict is None:
            sample_df_dict = mdl.download_sample(ckan=ckan, **kwargs)
        if return_sample:
            return sample_mdl, sample_df_dict
        else:
            return sample_mdl

    def info_request_package(self, ckan:CkanApi) -> CkanPackageInfo:
        pkg_info = ckan.get_package_info_or_request(package_name=self.package_name)
        self.package_attributes.id = pkg_info.id
        return pkg_info

    def info_request_full(self, ckan:CkanApi) -> Tuple[CkanPackageInfo, List[CkanResourceInfo]]:
        pkg_info = self.info_request_package(ckan)
        self.update_package_name_in_resources()
        self.init_resources_options_and_metadata(ckan, base_dir=self.get_base_dir())
        res_info = [resource_builder.resource_info_request(ckan) for resource_builder in self.resource_builders.values()]
        return pkg_info, res_info

    def get_base_dir(self, base_dir:str=None) -> str:
        """
        Returns the default base_dir if not specified. The base_dir is the location of the Excel workbook.
        If this was initialized from a dictionary, the current working directory will be used (cwd).

        :return:
        """
        if base_dir is None:
            if self.builder_source_file is not None:
                base_dir, _ = os.path.split(self.builder_source_file)
            else:
                base_dir = os.path.abspath(".")
        return base_dir

    def get_resources_base_dir(self, resources_base_dir:str) -> str:
        """
        This returns the base directory for the resource files.
        It is distinct from the base_dir and can be defined relative to the base_dir in the Excel workbook (see comment at the top of the class).

        :param resources_base_dir:
        :return:
        """
        if resources_base_dir is None:
            resources_base_dir = self._resources_base_dir
        return resources_base_dir

    def get_default_out_dir(self, out_dir:str, enforce_none:bool=False) -> str:
        """
        This returns the default download directory.

        :param out_dir:
        :return:
        """
        if out_dir is None and not enforce_none:
            out_dir = self._default_out_dir
        return out_dir

    def init_ckan(self, ckan:CkanApi=None, *, base_dir:str=None, set_owner_org:bool=False,
                  default_proxies:dict=None, proxies:Union[str,dict,ProxyConfig]=None) -> CkanApi:
        """
        Initialize the CKAN instance from the parameters defined in the "ckan" tab of the Excel workbook.

        :param ckan: 
        :param base_dir: 
        :param default_proxies: 
        :param set_owner_org: Option to set the owner_org of the CKAN instance.
        This can be problematic because it requires some requests as the proxies are not set.
        It can be omitted because it has no influence on the patch_request_package function.
        :return: 
        """
        base_dir = self.get_base_dir(base_dir)  # base_dir is necessary to find the API key file, if provided
        ckan = self.ckan_builder.init_ckan(base_dir, ckan=ckan, default_proxies=default_proxies,
                                           proxies=proxies)
        if set_owner_org and self.organization_name is not None:
            ckan.owner_org = self.get_owner_org(ckan)
        return ckan

    def get_or_query_package_id(self, ckan: CkanApi) -> str:
        package_info = ckan.get_package_info_or_request(self.package_name)
        self.package_attributes.id = package_info.id
        return package_info.id

    def list_resource_ids(self, ckan: CkanApi) -> List[str]:
        """
        List resource ids on CKAN server, following the order of the package builder

        :param ckan:
        :return:
        """
        self.update_package_name_in_resources()
        self.init_resources_options_and_metadata(ckan, base_dir=self.get_base_dir())
        mono_resource_names = {resource_name for resource_name, resource_builder in self.resource_builders.items() if not isinstance(resource_builder, BuilderMultiFile)}
        resource_ids = []
        for resource_builder in self.resource_builders.values():
            if not (isinstance(resource_builder, BuilderMultiFile)):
                resource_ids.append(resource_builder.get_or_query_resource_id(ckan))
            else:
                multi_resource_ids = resource_builder.list_remote_resource_ids(ckan, excluded_resource_names=mono_resource_names,
                                                                               cancel_if_present=False)
                resource_ids = resource_ids + multi_resource_ids
        np_resource_ids = np.array(resource_ids)
        _, I = np.unique(np_resource_ids, return_index=True)
        I.sort()
        np_resource_ids = np_resource_ids[I]
        resource_ids = np_resource_ids.tolist()
        return resource_ids

    def package_resource_reorder(self, ckan: CkanApi) -> None:
        """
        Apply the order of the resources defined in the Excel workbook.

        :param ckan: 
        :return: 
        """
        # OrderedDict ensures the order of resources is preserved
        package_id = self.get_or_query_package_id(ckan=ckan)
        resource_ids = self.list_resource_ids(ckan=ckan)
        ckan._api_package_resource_reorder(package_id=package_id, resource_ids=resource_ids)

    def remote_policy_check(self, ckan: CkanApi, policy:CkanPackageDataFormatPolicy=None,
                            *, buffer:Dict[str, List[DataPolicyError]]=None, raise_error:bool=False,
                            verbose:bool=None) -> bool:
        """
        Check the package defined by this builder against a data format policy, based on the information from the API.

        :param ckan:
        :param policy:
        :param buffer:
        :param raise_error:
        :param verbose:
        :return:
        """
        if policy is None:
            policy = self.ckan_builder.policy
        return ckan.policy_check(package_list=self.package_name, policy=policy, buffer=buffer,
                                 verbose=verbose, raise_error=raise_error)

    def local_policy_check(self, policy:CkanPackageDataFormatPolicy=None,
                           *, buffer:Dict[str, List[DataPolicyError]]=None, raise_error:bool=False,
                           verbose:bool=True) -> bool:
        """
        Check if the package builder respects a data format policy (only on local definition).

        :return:
        """
        if policy is None:
            policy = self.ckan_builder.policy
        if policy is None:
            # no policy loaded at all
            return True
        package_info = self.to_ckan_package_info(check_id=False)
        package_buffer: List[DataPolicyError] = []
        success = policy.policy_check_package(package_info, display_message=verbose,
                                              package_buffer=package_buffer, raise_error=raise_error)
        if buffer is not None:
            buffer[package_info.name] = package_buffer
        if verbose:
            print(f"Data format policy {policy.label} success: {success}")
        return success





#!python3
# -*- coding: utf-8 -*-
"""
The basic file format for DataStore: CSV
"""
from typing import Union, Dict
import io

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.auxiliary.list_records import ListRecords
from ckanapi_harvesters.auxiliary.ckan_auxiliary import df_download_to_csv_kwargs
from ckanapi_harvesters.harvesters.file_formats.file_format_abc import FileFormatABC



class CsvFileFormat(FileFormatABC):
    default_read_kwargs = dict(dtype=str, keep_default_na=False, sep=None, engine='python')  # read, (sep=None, engine='python') enables the detection of the column separator
    default_write_kwargs = df_download_to_csv_kwargs  # write

    # read -------------------
    def read_by_chunks_allowed(self) -> bool:
        return True

    def _get_read_kwargs(self, allow_chunks:bool=True) -> dict:
        kwargs = super()._get_read_kwargs(allow_chunks=allow_chunks)
        if self.read_by_chunks_enabled(allow_chunks=allow_chunks):
            kwargs["chunksize"] = self.chunk_size
        else:
            kwargs["chunksize"] = None
        return kwargs

    def read_file(self, file_path: str, fields: Union[Dict[str, CkanField],None], allow_chunks:bool=True) -> Union[pd.DataFrame, ListRecords]:
        read_kwargs = self._get_read_kwargs(allow_chunks=allow_chunks)
        return pd.read_csv(file_path, **read_kwargs)

    def read_buffer_full(self, buffer: io.StringIO, fields: Union[Dict[str, CkanField],None]) -> Union[pd.DataFrame, ListRecords]:
        read_kwargs = self._get_read_kwargs(allow_chunks=False)
        return pd.read_csv(buffer, **read_kwargs)

    # write ------------------
    def write_file(self, df: pd.DataFrame, file_path: str, fields: Union[Dict[str, CkanField],None]) -> None:
        write_kwargs = self._get_write_kwargs()
        df.to_csv(file_path, index=False, **write_kwargs)

    def write_in_memory(self, df: pd.DataFrame, fields: Union[Dict[str, CkanField],None]) -> bytes:
        with io.StringIO() as stream:
            write_kwargs = self._get_write_kwargs()
            df.to_csv(stream, index=False, **write_kwargs)
            return stream.getvalue().encode("utf8")

    def append_allowed(self) -> bool:
        return True

    def append_file(self, df: Union[pd.DataFrame, ListRecords], file_path: str,
                    fields: Union[Dict[str, CkanField], None]) -> None:
        write_kwargs = self._get_write_kwargs()
        df.to_csv(file_path, index=False, mode='a', **write_kwargs)

    def append_in_memory(self, stream: bytes, df: Union[pd.DataFrame, ListRecords], fields: Union[Dict[str, CkanField],None]) -> bytes:
        with io.StringIO(stream.decode("utf8")) as string_stream:
            write_kwargs = self._get_write_kwargs()
            df.to_csv(string_stream, index=False, mode='a', **write_kwargs)
            return string_stream.getvalue().encode("utf8")

    # misc ------------------
    def copy(self, dest=None):
        if dest is None:
            dest = CsvFileFormat(self.options_string, read_kwargs=self.read_kwargs, write_kwargs=self.write_kwargs)
        super().copy(dest=dest)
        return dest


if __name__ == '__main__':
    sample_instance = CsvFileFormat("--read-kwargs compression=gzip header=10")
    print("File format reader CLI-format options:")
    sample_instance.print_help_cli()




#!python3
# -*- coding: utf-8 -*-
"""
Tests to perform after the example package was uploaded
"""
import os
import json

import pandas as pd

from ckanapi_harvesters.auxiliary import CkanMap
from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.ckan_api import CkanApi

from ckanapi_harvesters.builder.example import example_package_xls, example_package_resources_dir


def traces_chunks_generator(chunksize:int):
    traces_file = os.path.join(example_package_resources_dir, "traces", "trace_000.csv")
    return pd.read_csv(traces_file, chunksize=chunksize)


def run(ckan:CkanApi = None):
    BuilderPackage.unlock_external_code_execution()  # comment to test if the safety feature is enabled

    mdl = BuilderPackage.from_excel(example_package_xls)
    ckan = mdl.init_ckan(ckan)
    ckan.initialize_from_cli_args()
    ckan.input_missing_info(input_args_if_necessary=True, input_owner_org=True)
    ckan.set_verbosity(True)

    # Test sending the traces in a generator mode
    resource_id = mdl.get_or_query_resource_id(ckan, resource_name="traces.csv")
    with traces_chunks_generator(chunksize=6) as records_generator:
        ckan.datastore_upsert_generator(records_generator, resource_id=resource_id, request_threshold=15, offset=12)

    # Test re-encoding the Excel file from the loaded model
    example_package_xls_reencoded = os.path.abspath("builder_package_example-reencoded.xlsx")
    mdl.to_excel(example_package_xls_reencoded)

    # map package
    ckan.map_resources(mdl.package_name, datastore_info=True, license_list=True)
    map_init = ckan.map.copy()
    ckan.purge(purge_map=True)

    # Test using the model to update CKAN map (update_ckan_map)
    mdl.info_request_full(ckan)
    map_queried = ckan.map.copy()
    ckan.purge(purge_map=True)
    mdl.update_ckan_map(ckan)  # use known resource ids to fill ckan.map (at your own risks)
    map_from_mdl = ckan.map.copy()

    # Test saving the map to a dictionary
    dict_map_queried = map_queried.to_dict()
    map_queried_from_dict = CkanMap.from_dict(dict_map_queried)

    dict_map_from_mdl = map_from_mdl.to_dict()
    map_from_mdl_from_dict = CkanMap.from_dict(dict_map_from_mdl)

    # test the function that recreates the Excel file from the online information
    ckan.purge(purge_map=True)
    mdl_api = BuilderPackage.from_ckan(ckan, mdl.package_name)
    example_package_xls_from_api = os.path.abspath("builder_package_example-from-api.xlsx")
    mdl_api.to_excel(example_package_xls_from_api)

    # export mdl to dict and re-import
    base_dir = os.path.abspath(".")
    mdl_dict = mdl.to_dict(base_dir=base_dir)
    mdl_from_dict = BuilderPackage.from_dict(mdl_dict, base_dir=base_dir)
    mdl_dict_bis = mdl_from_dict.to_dict(base_dir=base_dir)
    assert(mdl_dict == mdl_dict_bis)

    # test json serialization
    example_package_json_reencoded = os.path.abspath("builder_package_example-reencoded.json")
    with open(example_package_json_reencoded, "w") as json_file:
        json.dump(mdl_dict, json_file, indent=4)
        json_file.close()

    # test copy constructors
    ckan_copy = ckan.copy()
    mdl_copy = mdl.copy()
    mdl_copy_dict = mdl_copy.to_dict(base_dir=base_dir)
    assert(mdl_dict == mdl_copy_dict)

    print("Tests done.")


if __name__ == '__main__':
    ckan = CkanApi(None)
    ckan.initialize_from_cli_args()
    run(ckan)




#!python3
# -*- coding: utf-8 -*-
"""
Data model to represent a CKAN database architecture
"""
from typing import Iterable
from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel

## Exceptions ------------------
class EmptyPackageNameException(RuntimeError):
    def __init__(self):
        super().__init__("Run-time error: the attribute package_name cannot be empty")

class MissingDataStoreColumnsSheet(Exception):
    def __init__(self, resource_name: str, columns_sheet_name: str):
        super().__init__(f"DataStore {resource_name} specifies a DataStore Columns Sheet '{columns_sheet_name}' but it was not found in the spreadsheet.")

class MissingDataStoreInfoError(Exception):
    def __init__(self):
        super().__init__("DataStore info must be requested to initiate resource builder. Use option datastore_info=True for the map_resources function.")

class RequiredDataFrameFieldsError(Exception):
    def __init__(self, missing_fields:Iterable[str]):
        super().__init__("The following fields are required but absent from the sample DataFrame: {}".format(", ".join(missing_fields)))

class UnsupportedBuilderVersionError(Exception):
    def __init__(self, file_version):
        super().__init__(f"Version error: package builder version {file_version} is not supported")

class ResourceFileNotExistMessage(ContextErrorLevelMessage):
    def __init__(self, resource_name:str, error_level:ErrorLevel, specific_message: str):
        super().__init__(f"Resource {resource_name}", error_level, specific_message)

class GroupByError(Exception):
    pass



#!python3
# -*- coding: utf-8 -*-
"""
Package to make reports on the CKAN database.
"""

from . import admin_report

# usage shortcuts
from ckanapi_harvesters.reports.admin_report import CkanAdminReport





#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to convert formats between database and local files.
"""
from typing import Dict, List, Callable, Any, Union, Set
import copy
import os

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise
from ckanapi_harvesters.auxiliary.ckan_errors import MissingCodeFileError
from ckanapi_harvesters.auxiliary.external_code_import import PythonUserCode
from ckanapi_harvesters.auxiliary.list_records import ListRecords, GeneralDataFrame
from ckanapi_harvesters.auxiliary.ckan_model import CkanField


class DataSchemeConversion:
    def __init__(self, *, df_upload_fun:Callable[[GeneralDataFrame,Any], GeneralDataFrame] = None,
                 df_download_fun:Callable[[GeneralDataFrame,Any], GeneralDataFrame] = None):
        """
        Class to convert between local data formats and database formats

        :param df_upload_fun:
        :param df_download_fun:
        """
        self.df_upload_fun:Union[Callable[[GeneralDataFrame, Any], GeneralDataFrame], None] = df_upload_fun
        self.df_download_fun:Union[Callable[[GeneralDataFrame, Any], GeneralDataFrame], None] = df_download_fun
        self.upload_index_column: str = ""
        self.source_file_column: str = ""

    def copy(self):
        return copy.deepcopy(self)

    def df_upload_alter(self, df_local: Union[pd.DataFrame, List[dict], Any], *,
                        total_lines_read: int, fields:Dict[str, CkanField], file_query:str,
                        mapper_kwargs:dict=None, **kwargs) -> Union[pd.DataFrame, ListRecords]:
        """
        Apply used-defined df_upload_fun if present

        :param df_local: the DataFrame to upload
        :param total_lines_read: total number of lines read, including the current DataFrame
        :param fields: the known fields metadata.
        :param file_query: the name of the file the data originates from (or query)
        :param mapper_kwargs: extra arguments passed to df_upload_fun
        :return: the DataFrame ready for upload, converted in the format of the database
        """
        if mapper_kwargs is None: mapper_kwargs = {}
        mapper_kwargs["file_query"] = file_query
        mapper_kwargs["file_name"] = file_query  # retro-compatible argument (deprecated)
        mapper_kwargs["fields"] = fields
        mapper_kwargs["total_lines_read"] = total_lines_read
        if file_query is not None and (isinstance(df_local, pd.DataFrame) or isinstance(df_local, ListRecords)):
            df_local.attrs["source"] = file_query
        if self.upload_index_column:
            # insert an extra column keeping track of the last read line (default primary key)
            index_offset = total_lines_read - len(df_local)
            if isinstance(df_local, pd.DataFrame):
                index_offset -= df_local.index[0]  # index of DataFrame in file, not 0 if the file is read by chunks
                assert_or_raise(not(self.upload_index_column in df_local.keys()), KeyError(f"{self.upload_index_column} already exists"))
                df_local[self.upload_index_column] = df_local.index + index_offset
            else:
                for index, line in enumerate(df_local):
                    assert_or_raise(not(self.upload_index_column in line.keys()), KeyError(f"{self.upload_index_column} already exists"))
                    line[self.upload_index_column] = index + index_offset
        if self.source_file_column:
            _, file_name = os.path.split(file_query)
            df_local[self.source_file_column] = file_name
        df_database = df_local
        if self.df_upload_fun is not None:
            # df_database = df_local.copy()  # unnecessary copy
            df_upload_fun = self.df_upload_fun
            df_database = df_upload_fun(df_database, **mapper_kwargs, **kwargs)
        if not isinstance(df_database, pd.DataFrame):
            if isinstance(df_database, ListRecords):
                pass  # also accept ListRecords (List[dict])
            elif isinstance(df_database, list):
                df_database = ListRecords(df_database)
            elif self.df_upload_fun is None:
                raise TypeError("No upload function was defined to convert the data format to a DataFrame")
            else:
                raise TypeError("df_upload_fun must return a DataFrame")
        return df_database

    def df_download_alter(self, df_database:Union[pd.DataFrame, List[dict], Any], file_query:dict=None,
                          fields:Dict[str, CkanField]=None, mapper_kwargs:dict=None, **kwargs) -> Union[pd.DataFrame, ListRecords]:
        """
        Apply used-defined df_download_fun if present.
        df_download_fun should be the reverse function of df_upload_fun

        :param df_database: the downloaded dataframe from the database
        :return: the dataframe ready to save, converted in the local format
        """
        if mapper_kwargs is None: mapper_kwargs = {}
        mapper_kwargs["file_query"] = file_query
        if file_query is not None:
            df_database.attrs["query"] = file_query
        df_local = df_database
        if self.df_download_fun is not None:
            # df_local = df_database.copy()  # unnecessary copy
            df_download_fun = self.df_download_fun
            df_local = df_download_fun(df_local, fields=fields, **mapper_kwargs, **kwargs)
        if not isinstance(df_local, pd.DataFrame):
            if isinstance(df_local, ListRecords):
                pass  # also accept ListRecords (List[dict])
            elif isinstance(df_local, list):
                df_local = ListRecords(df_local)
            elif self.df_download_fun is None:
                raise TypeError("No download function was defined to convert the received DataFrame")
            else:
                raise TypeError("df_download_fun must return a DataFrame")
        return df_local

    def _connect_aux_functions(self, module: PythonUserCode, aux_upload_fun_name:str, aux_download_fun_name:str) -> None:
        if (aux_upload_fun_name or aux_download_fun_name) and module is None:
            raise MissingCodeFileError()
        if aux_upload_fun_name:
            self.df_upload_fun = module.function_pointer(aux_upload_fun_name)
        if aux_download_fun_name:
            self.df_download_fun = module.function_pointer(aux_download_fun_name)

    def get_necessary_fields(self) -> Set[str]:
        return set()




#!python3
# -*- coding: utf-8 -*-
"""
The basic file format for DataStore: CSV
"""
from typing import Union, Dict, Callable, Any
import io
import argparse

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_model import CkanField, CkanResourceInfo
from ckanapi_harvesters.auxiliary.list_records import ListRecords, GeneralDataFrame
from ckanapi_harvesters.auxiliary.ckan_errors import MissingCodeFileError, MissingIOFunctionError
from ckanapi_harvesters.auxiliary.external_code_import import PythonUserCode
from ckanapi_harvesters.harvesters.file_formats.file_format_abc import FileFormatABC


class UserFileFormat(FileFormatABC):
    def __init__(self, options_string: str, *, df_read_fun:Callable[[Any], GeneralDataFrame] = None,
                 df_write_fun:Callable[[GeneralDataFrame, Any], Any] = None,
                 read_kwargs:dict=None, write_kwargs:dict=None) -> None:
        super().__init__(options_string=options_string, read_kwargs=read_kwargs, write_kwargs=write_kwargs)
        self.df_read_fun:Union[Callable[[Any], GeneralDataFrame], None] = df_read_fun
        self.df_write_fun:Union[Callable[[pd.DataFrame, Any], pd.DataFrame], None] = df_write_fun
        self.option_append_allowed: bool = False

    @staticmethod
    def _setup_cli_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        parser = FileFormatABC._setup_cli_parser(parser)
        parser.add_argument("--allow-append",
                            help="Option to signal the append mode is available for the write function", action="store_true", default=False)
        return parser

    def _apply_arguments(self, args: argparse.Namespace, extra_args: list):
        super()._apply_arguments(args, extra_args)
        self.option_append_allowed = args.allow_append

    def _connect_aux_functions(self, module: PythonUserCode, aux_read_fun_name:str, aux_write_fun_name:str) -> None:
        if (aux_read_fun_name or aux_write_fun_name) and module is None:
            raise MissingCodeFileError()
        if aux_read_fun_name:
            self.df_read_fun = module.function_pointer(aux_read_fun_name)
        if aux_write_fun_name:
            self.df_write_fun = module.function_pointer(aux_write_fun_name)

    # read -------------------
    def read_by_chunks_allowed(self) -> bool:
        return True

    def read_file(self, file_path: str, fields: Union[Dict[str, CkanField],None], allow_chunks:bool=True) -> Union[pd.DataFrame, ListRecords]:
        if self.df_read_fun is None:
            raise MissingIOFunctionError("Read function")
        read_kwargs = self._get_read_kwargs(allow_chunks=allow_chunks)
        self.resource_attributes_from_file = None
        output = self.df_read_fun(file_path, fields=fields, allow_chunks=self.read_by_chunks_enabled(allow_chunks=allow_chunks), params=self, **read_kwargs)
        if isinstance(output, tuple):
            self.resource_attributes_from_file = output[1]
            return output[0]
        else:
            return output

    def read_buffer_full(self, buffer: io.StringIO, fields: Union[Dict[str, CkanField],None]) -> Union[pd.DataFrame, ListRecords]:
        if self.df_read_fun is None:
            raise MissingIOFunctionError("Read function")
        read_kwargs = self._get_read_kwargs(allow_chunks=False)
        self.resource_attributes_from_file = None
        output = self.df_read_fun(buffer, fields=fields, allow_chunks=False, params=self, **read_kwargs)
        if isinstance(output, tuple):
            assert(isinstance(output[1], CkanResourceInfo))
            self.resource_attributes_from_file = output[1]
            return output[0]
        else:
            return output

    # write ------------------
    def write_file(self, df: pd.DataFrame, file_path: str, fields: Union[Dict[str, CkanField],None]) -> None:
        if self.df_write_fun is None:
            raise MissingIOFunctionError("Write function")
        write_kwargs = self._get_write_kwargs()
        self.df_write_fun(df, file_path, append=False, fields=fields, params=self, **write_kwargs)

    def append_allowed(self) -> bool:
        return self.option_append_allowed

    def append_file(self, df: Union[pd.DataFrame, ListRecords], file_path: str,
                    fields: Union[Dict[str, CkanField], None]) -> None:
        if self.df_write_fun is None:
            raise MissingIOFunctionError("Write function")
        write_kwargs = self._get_write_kwargs()
        self.df_write_fun(df, file_path, append=True, fields=fields, params=self, **write_kwargs)

    def write_in_memory(self, df: pd.DataFrame, fields: Union[Dict[str, CkanField],None]) -> bytes:
        if self.df_write_fun is None:
            raise MissingIOFunctionError("Write function")
        with io.StringIO() as buffer:
            write_kwargs = self._get_write_kwargs()
            self.df_write_fun(df, buffer, append=False, fields=fields, params=self, **write_kwargs)
            return buffer.getvalue().encode("utf8")

    def append_in_memory(self, stream: bytes, df: Union[pd.DataFrame, ListRecords], fields: Union[Dict[str, CkanField],None]) -> bytes:
        if self.df_write_fun is None:
            raise MissingIOFunctionError("Write function")
        with io.StringIO(stream.decode("utf8")) as string_stream:
            write_kwargs = self._get_write_kwargs()
            self.df_write_fun(df, string_stream, append=True, fields=fields, params=self, **write_kwargs)
            return string_stream.getvalue().encode("utf8")

    # misc ------------------
    def copy(self, dest=None):
        if dest is None:
            dest = UserFileFormat(self.options_string, read_kwargs=self.read_kwargs, write_kwargs=self.write_kwargs,
                                  df_read_fun=self.df_read_fun, df_write_fun=self.df_write_fun)
        super().copy(dest=dest)
        return dest




#!python3
# -*- coding: utf-8 -*-
"""
Data format policy representation and enforcing for lists of values such as tags
"""
from typing import List, Union, Set

from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.policies.data_format_policy_errors import DataPolicyError, ErrorLevel, _policy_msg
from ckanapi_harvesters.policies.data_format_policy_defs import ListChoiceMode
from ckanapi_harvesters.policies.data_format_policy_defs import StringValueSpecification
from ckanapi_harvesters.policies.data_format_policy_abc import DataPolicyElementABC


extra_group_name = "extra"


class ValueListPolicy(DataPolicyElementABC):
    _group_type_str = "group"

    def __init__(self, list_specs:List[StringValueSpecification]=None, group_name:str=None,
                 value_select:ListChoiceMode=ListChoiceMode.Any,
                 mandatory:bool=False, error_level:ErrorLevel=ErrorLevel.Information):
        super().__init__(mandatory=mandatory, error_level=error_level)
        if list_specs is None:
            list_specs = []
        self.list_specs:List[StringValueSpecification] = list_specs
        self.value_select: ListChoiceMode = value_select
        self.group_name: str = group_name

    def to_dict(self) -> dict:
        d = {}
        if self.group_name:
            d["group_name"] = self.group_name
        d.update(super().to_dict())
        d.update({"values": [spec.to_dict() for spec in self.list_specs],
                  "value_select": self.value_select.name})
        return d

    def list_specs_str(self) -> List[str]:
        return [value_spec.value for value_spec in self.list_specs]

    @staticmethod
    def from_dict(d:dict) -> "ValueListPolicy":
        obj = ValueListPolicy()
        obj._load_from_dict(d)
        return obj

    def _load_from_dict(self, d:dict):
        super()._load_from_dict(d)
        self.group_name = d["group_name"] if len(d["group_name"]) > 0 else None
        self.list_specs = [StringValueSpecification.from_dict(value)  for value in d["values"]]
        self.value_select = ListChoiceMode.from_str(d["value_select"]) if "value_select" in d.keys() else ListChoiceMode.Any

    def enforce(self, values: Union[str, List[str]], *, context:str=None,
                verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        if self.group_name is not None:
            context = context + " / " + self._group_type_str + " " + self.group_name
        success = True
        spec = [tag_spec.value for tag_spec in self.list_specs]
        if values is None:
            values = []
        elif isinstance(values, str):
            values = values.split(ckan_tags_sep)
        values = list(set(values).intersection(set(spec)))
        msg = None
        value_context = context + " / value '" + ','.join(values).join(values) + "'"
        if (self.value_select == ListChoiceMode.MaxOne and len(values) > 1):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"Too many values for value list group '{self.group_name}'. Max one value is admitted within {spec}.")
        if (self.value_select == ListChoiceMode.NoExtra and len(values) > 0):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"Too many values for value list group '{self.group_name}'. No values can be selected for this group ({spec}).")
        if (self.value_select == ListChoiceMode.MandatoryOne and not len(values) == 1):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"Exactly one value must be present for value list group '{self.group_name}' ({spec}).")
        if (self.value_select == ListChoiceMode.MandatoryMulti and not len(values) < 1):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"At least one value must be present for value list group '{self.group_name}' ({spec}).")
        if not success:
            _policy_msg(msg, error_level=self.error_level, buffer=buffer, verbose=verbose)
        if len(spec) > 0:
            for tag in values:
                success &= self._enforce_unit_string(tag, spec, context=context, verbose=verbose, buffer=buffer)
        return success


class ExtraValueListPolicy(ValueListPolicy):
    def __init__(self, list_specs:List[StringValueSpecification]=None,
                 value_select:ListChoiceMode=ListChoiceMode.Any,
                 mandatory:bool=False, error_level:ErrorLevel=ErrorLevel.Information):
        super().__init__(list_specs=list_specs, group_name=extra_group_name,
                         value_select=value_select, mandatory=mandatory, error_level=error_level)

    @staticmethod
    def from_ValueListPolicy(value: ValueListPolicy) -> "ExtraValueListPolicy":
        obj = ExtraValueListPolicy()
        obj.__dict__.update(value.__dict__)
        return obj

    @staticmethod
    def from_dict(d:dict) -> "ExtraValueListPolicy":
        obj = ExtraValueListPolicy()
        obj._load_from_dict(d)
        return obj

    def enforce(self, values: Union[str, List[str]], *, context:str=None,
                verbose: bool = True, buffer:List[DataPolicyError]=None, extra_spec_rm:Set[str]=None) -> bool:
        if self.group_name is not None:
            context = context + " / group " + self.group_name
        success = True
        spec = [tag_spec.value for tag_spec in self.list_specs]
        if values is None:
            values = []
        elif isinstance(values, str):
            values = values.split(ckan_tags_sep)
        values = list(set(values) - extra_spec_rm)
        msg = None
        value_context = context + " / value '" + ','.join(values).join(values) + "'"
        if (self.value_select == ListChoiceMode.MaxOne and len(values) > 1):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"Too many values for value list group '{self.group_name}'. Max one value is admitted within {spec}.")
        if (self.value_select == ListChoiceMode.NoExtra and len(values) > 0):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"Too many values for value list group '{self.group_name}'. No values can be selected for this group ({spec}). Admitted values from other groups: {extra_spec_rm}")
        if (self.value_select == ListChoiceMode.MandatoryOne and not len(values) == 1):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"Exactly one value must be present for value list group '{self.group_name}' ({spec}).")
        if (self.value_select == ListChoiceMode.MandatoryMulti and not len(values) < 1):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"At least one value must be present for value list group '{self.group_name}' ({spec}).")
        if not success:
            _policy_msg(msg, error_level=self.error_level, buffer=buffer, verbose=verbose)
        if len(spec) > 0:
            for tag in values:
                success &= self._enforce_unit_string(tag, spec, context=context, verbose=verbose, buffer=buffer)
        return success


class GroupedValueListPolicy(DataPolicyElementABC):
    def __init__(self, value_group_specs:List[ValueListPolicy]=None,
                 extra_values:ExtraValueListPolicy=None,
                 mandatory:bool=False, error_level:ErrorLevel=ErrorLevel.Information):
        super().__init__(mandatory=mandatory, error_level=error_level)
        if value_group_specs is None:
            value_group_specs = []
        self.value_group_specs:List[ValueListPolicy] = value_group_specs
        self.extra_values_spec:ExtraValueListPolicy = extra_values
        self._extract_extra_values()

    def _extract_extra_values(self):
        i_rm = []
        extra_values = self.extra_values_spec
        for i, value_group_spec in enumerate(self.value_group_specs):
            if value_group_spec.group_name.lower() == extra_group_name.lower():
                assert(extra_values is None)
                extra_values = ExtraValueListPolicy.from_ValueListPolicy(value_group_spec)
                i_rm.append(i)
        for i in reversed(i_rm):
            self.value_group_specs.pop(i)
        if extra_values is not None:
            self.extra_values_spec:ExtraValueListPolicy = extra_values

    def to_dict(self) -> dict:
        d = super().to_dict()
        if self.extra_values_spec is not None:
            self.extra_values_spec.group_name = extra_group_name
            extra_values_dict = [self.extra_values_spec.to_dict()]
        else:
            extra_values_dict = []
        d.update({"groups": [spec.to_dict() for spec in self.value_group_specs] + extra_values_dict})
        return d

    @staticmethod
    def from_dict(d:dict) -> "GroupedValueListPolicy":
        obj = GroupedValueListPolicy()
        obj._load_from_dict(d)
        return obj

    def _load_from_dict(self, d:dict, child_cls:type=None):
        super()._load_from_dict(d)
        if child_cls is None:
            child_cls = ValueListPolicy
        self.value_group_specs = [child_cls.from_dict(group_spec) for group_spec in d["groups"]]
        self.extra_values_spec = None
        self._extract_extra_values()

    def enforce(self, values: Union[str, List[str]], *, context:str=None, verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        success = True
        extra_spec_rm = set()
        for value_group_spec in self.value_group_specs:
            if not value_group_spec.group_name == extra_group_name.lower():
                extra_spec_rm = extra_spec_rm.union({tag_spec.value for tag_spec in value_group_spec.list_specs})
        for value_group_spec in self.value_group_specs:
            success &= value_group_spec.enforce(values, context=context, verbose=verbose, buffer=buffer)
        if self.extra_values_spec is not None:
            self.extra_values_spec.group_name = extra_group_name
            success &= self.extra_values_spec.enforce(values, context=context, verbose=verbose, buffer=buffer, extra_spec_rm=extra_spec_rm)
        return success


class SingleValueListPolicy(DataPolicyElementABC):
    def __init__(self, base_list:ValueListPolicy=None, extra_values:ListChoiceMode=ListChoiceMode.Any, mandatory:bool=False):
        super().__init__(mandatory=mandatory)
        self.base_list: GroupedValueListPolicy = GroupedValueListPolicy(extra_values=ExtraValueListPolicy(value_select=extra_values))
        self.update_base_list(base_list)

    def to_dict(self) -> dict:
        return self.base_list.to_dict()

    @staticmethod
    def from_dict(d:dict) -> "SingleValueListPolicy":
        obj = SingleValueListPolicy()
        obj._load_from_dict(d)
        return obj

    def _load_from_dict(self, d:dict):
        # super()._load_from_dict(d)
        self.base_list._load_from_dict(d)

    def update_base_list(self, base_list:ValueListPolicy):
        if base_list is not None:
            base_list.group_name = "base"
            self.base_list.value_group_specs = [base_list]
        else:
            self.base_list.value_group_specs = []

    def enforce(self, values: Union[str, List[str]], *, context:str=None, verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        success = self.base_list.enforce(values, context=context, verbose=verbose, buffer=buffer)
        return success



#!python3
# -*- coding: utf-8 -*-
"""
Tests to perform after the example package was uploaded
"""
import os
import json

from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.policies.data_format_policy import CkanPackageDataFormatPolicy
from ckanapi_harvesters.policies.data_format_policy import (SingleValueListPolicy, ValueListPolicy, StringValueSpecification,
                                                            ListChoiceMode, CustomFieldsPolicy, CustomFieldSpecification,
                                                            ErrorLevel, DataPolicyError,
                                                            StringMatchMode, TagListPolicy, TagGroupsListPolicy)

from ckanapi_harvesters.builder.specific.configuration_builder import ConfigurationBuilder
from ckanapi_harvesters.builder.example import example_package_xls


enable_write = False  # be careful, setting this to True could erase your CKAN data format policy !!!


def run(ckan:CkanApi = None):
    BuilderPackage.unlock_external_code_execution()  # comment to test if the safety feature is enabled

    mdl = BuilderPackage.from_excel(example_package_xls)
    ckan = mdl.init_ckan(ckan)
    ckan.initialize_from_cli_args()
    ckan.input_missing_info(input_args_if_necessary=True, input_owner_org=True)
    ckan.set_verbosity(False)
    ckan.verbose_policy = True

    policy = CkanPackageDataFormatPolicy()
    policy.resource_format = SingleValueListPolicy(ValueListPolicy([StringValueSpecification("CSV")]), extra_values=ListChoiceMode.NoExtra)
    policy.package_custom_fields = CustomFieldsPolicy([
                                                        CustomFieldSpecification(key="New field", mandatory=True),
                                                        CustomFieldSpecification(key="Algorithm", values=["Random"], match_mode=StringMatchMode.Match, mandatory=True)], error_level=ErrorLevel.Error)
    policy.package_tags = TagGroupsListPolicy([TagListPolicy([StringValueSpecification("Test")], group_name="Vocabulary_Test")])
    policy.package_mandatory_attributes = {"description", "author"}
    ckan.policy = policy

    policy_dict = policy.to_dict(sets_as_lists=False)
    policy_from_dict = CkanPackageDataFormatPolicy.from_dict(policy_dict)
    policy_dict_bis = policy_from_dict.to_dict(sets_as_lists=False)
    assert(policy_dict == policy_dict_bis)

    # serialisation
    policy_dict = policy.to_dict(sets_as_lists=True)
    policy_json_file = os.path.abspath("policy_py.json")
    with open(policy_json_file, "w") as f:
        json.dump(policy_dict, f, indent=4)
    with open(policy_json_file, "r") as f:
        policy_json_dict = json.load(f)
    # policy_json = CkanPackageDataFormatPolicy.from_dict(policy_json_dict)
    policy_json = CkanPackageDataFormatPolicy.from_json(policy_json_file)
    assert(policy_dict == policy_json_dict)

    # test on local definition (offline)
    buffer = {}
    success = mdl.local_policy_check(policy, buffer=buffer)
    # test if an error is raised (this mode does not display all messages)
    try:
        success = mdl.local_policy_check(policy)
    except DataPolicyError as e:
        print("Exception: " + str(e))
        assert(not success)
    else:
        print(f"No exception / success={success}")

    # test on remote definition (CKAN API)
    print("Test on remote")
    buffer = {}
    ckan.map_resources(mdl.package_name)
    success = ckan.policy_check(buffer=buffer)

    ckan.set_verbosity(True)

    # update vocabularies: deprecated
    # if enable_write:
    #     ckan.vocabularies_clear()
    #     ckan.initiate_vocabularies_from_policy(policy, remove_others=True)

    # upload default policy
    config_ckan = ConfigurationBuilder(ckan, ckan.owner_org)
    if enable_write:
        config_ckan.patch_policy(ckan, policy, reduced_size=False)

    # download default policy
    # default_policy = ckan.load_default_policy(force=True)
    config_ckan.load_default_policy(ckan)

    # test updating the scores on a CKAN package
    policy_auto_store = policy.copy()
    policy_auto_store.output_custom_fields.set_metadata_policy_fields(score="Data Policy Score", report="Data Policy Report")
    ckan.policy_check("configuration", policy=policy_auto_store)

    # check all packages against policy
    ckan.owner_org = None
    print(" ")
    buffer = {}
    config_ckan.policy_check(ckan, policy=policy, buffer=buffer, verbose=True)

    print("Tests done.")


if __name__ == '__main__':
    ckan = CkanApi(None)
    ckan.initialize_from_cli_args()
    run(ckan)




#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements the basic resources. See builder_datastore for specific functions to initiate datastores.
"""
from typing import Any, Generator, Union, Set, List, Dict, Tuple
import os
import requests
import copy

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, find_duplicates
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.auxiliary.ckan_errors import DuplicateNameError
from ckanapi_harvesters.auxiliary.path import resolve_rel_path, glob_rm_glob
from ckanapi_harvesters.auxiliary.list_records import ListRecords, GeneralDataFrame
from ckanapi_harvesters.builder.builder_resource_datastore import BuilderDataStoreABC
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC
from ckanapi_harvesters.harvesters.file_formats.file_format_init import FileFormatABC, init_file_format_datastore
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo
from ckanapi_harvesters.auxiliary.ckan_progress_callbacks import CkanCallbackLevel
from ckanapi_harvesters.builder.builder_field import BuilderField
from ckanapi_harvesters.builder.builder_resource_datastore_file import BuilderDataStoreFile
from ckanapi_harvesters.builder.builder_resource_multi_abc import FileChunkDataFrame
from ckanapi_harvesters.builder.builder_resource_multi_file import BuilderMultiFile


class BuilderMultiDataStore(BuilderMultiFile, BuilderDataStoreABC):
    def __init__(self, *, parent, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None):
        super().__init__(parent=parent, name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        self.field_builders: Union[Dict[str, BuilderField],None] = None
        self.field_builders_user: Union[Dict[str, BuilderField],None] = None
        self.field_builders_data_source: Union[Dict[str, BuilderField],None] = None
        self.primary_key: Union[List[str],None] = None
        self.indexes: Union[List[str],None] = None
        self.aux_upload_fun_name:str = ""
        self.aux_download_fun_name:str = ""
        self.aux_read_fun_name:str = ""
        self.aux_write_fun_name:str = ""
        self.data_cleaner_upload:Union[CkanDataCleanerABC,None] = None
        self.local_file_format: FileFormatABC = init_file_format_datastore(self.resource_attributes_user.format, self.options_string, self.aux_read_fun_name, self.aux_write_fun_name)
        self.read_line_counter: int = 0

    def copy(self, *, dest=None, parent=None):
        if dest is None:
            dest = BuilderMultiDataStore(parent=self.parent_package)
        super().copy(dest=dest, parent=parent)
        dest.field_builders = copy.deepcopy(self.field_builders)
        dest.field_builders_user = copy.deepcopy(self.field_builders_user)
        dest.field_builders_data_source = copy.deepcopy(self.field_builders_data_source)
        dest.primary_key = copy.deepcopy(self.primary_key)
        dest.indexes = copy.deepcopy(self.indexes)
        dest.aux_upload_fun_name = self.aux_upload_fun_name
        dest.aux_download_fun_name = self.aux_download_fun_name
        dest.local_file_format = self.local_file_format.copy()
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        primary_keys_string: str = _string_from_element(row["primary key"])
        indexes_string: str = _string_from_element(row["indexes"])
        if primary_keys_string is not None:
            if primary_keys_string.lower() == "none":
                self.primary_key = []
            else:
                self.primary_key = [field.strip() for field in primary_keys_string.split(ckan_tags_sep)]
        if indexes_string is not None:
            if indexes_string.lower() == "none":
                self.indexes = []
            else:
                self.indexes = [field.strip() for field in indexes_string.split(ckan_tags_sep)]
        if "upload function" in row.keys():
            self.aux_upload_fun_name: str = _string_from_element(row["upload function"], empty_value="")
        if "download function" in row.keys():
            self.aux_download_fun_name: str = _string_from_element(row["download function"], empty_value="")

    def _load_fields_df(self, fields_df: pd.DataFrame):
        fields_df.columns = fields_df.columns.map(str.lower)
        fields_df.columns = fields_df.columns.map(str.strip)
        self.field_builders_user = {}
        for row_loc, row in fields_df.iterrows():
            field_builder = BuilderField()
            field_builder._load_from_df_row(row=row)
            self.field_builders_user[field_builder.name] = field_builder

    def _check_field_duplicates(self):
        duplicates = find_duplicates([field_builder.name for field_builder in self.field_builders_user.values()])
        if len(duplicates) > 0:
            raise DuplicateNameError("Field", duplicates)

    def _get_fields_dict(self) -> Dict[str, dict]:
        self._check_field_duplicates()
        if self.field_builders_user is not None:
            fields_dict = {field_builder.name: field_builder._to_dict() for field_builder in self.field_builders_user.values()}
        else:
            fields_dict = None
        return fields_dict

    def _get_fields_df(self) -> Union[pd.DataFrame,None]:
        if self.field_builders_user is not None:
            fields_dict_list = [value for value in self._get_fields_dict().values()]
            fields_df = pd.DataFrame.from_records(fields_dict_list)
            return fields_df
        else:
            return None

    @staticmethod
    def resource_mode_str() -> str:
        return "MultiDataStore"

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["Primary key"] = ckan_tags_sep.join(self.primary_key) if self.primary_key else ""
        d["Indexes"] = ckan_tags_sep.join(self.indexes) if self.indexes is not None else ""
        return d

    def _data_store_builder_of_file(self, file_path:str) -> Tuple[BuilderDataStoreFile, str]:
        file_dir, file_name = os.path.split(file_path)
        ds_builder = BuilderDataStoreFile(parent=self.parent_package, name=file_name, description=self.resource_attributes_user.description, download_url=self.download_url,
                                          format=self.resource_attributes_user.format, file_name=file_name)
        ds_builder.field_builders = self.field_builders
        ds_builder.primary_key = self.primary_key
        ds_builder.indexes = self.indexes
        ds_builder.package_name = self.package_name
        ds_builder.aux_upload_fun_name = self.aux_upload_fun_name
        ds_builder.aux_download_fun_name = self.aux_download_fun_name
        ds_builder.aliases = None
        ds_builder.data_cleaner_upload = self.data_cleaner_upload
        ds_builder.local_file_format = self.local_file_format
        ds_builder.process_level = CkanCallbackLevel.MultiFileResource  # file of a multi-file resource
        return ds_builder, file_dir


    ## Upload ----------------
    def get_local_df_chunk_generator(self, resources_base_dir:str, ckan:CkanApi, excluded_files:Set[str]=None,
                                     allow_chunks:bool=True, **kwargs) -> Generator[FileChunkDataFrame, None, None]:
        self.list_local_files(resources_base_dir=resources_base_dir)
        for file_index, file_name in enumerate(self.local_file_list):
            self.file_semaphore.acquire()
            self.read_line_counter = 0
            self.file_semaphore.release()
            df_file = self.local_file_format.read_file(file_name, self.field_builders, allow_chunks=allow_chunks)
            if file_index == 0:
                self._merge_resource_attributes_from_file()
            if isinstance(df_file, pd.DataFrame) or isinstance(df_file, ListRecords):
                self.file_semaphore.acquire()
                self.read_line_counter += len(df_file)
                line_counter = self.read_line_counter
                self.file_semaphore.release()
                yield FileChunkDataFrame(df_file, file_name, file_index, 0, 0, line_counter)
            else:  # iterator
                with df_file as df_generator:
                    if hasattr(df_file, "handles"):
                        file_handle = df_file.handles.handle
                    else:
                        file_handle = None
                    previous_file_position = 0
                    # for chunk_index, df in enumerate(df_generator):
                    chunk_index = 0
                    file_position = 0
                    while True:
                        self.file_semaphore.acquire()
                        if file_handle is not None:
                            file_position = file_handle.buffer.tell()  # approximative position in file
                        else:
                            file_position = 0  # no position tracking available
                        try:
                            df = next(df_generator)
                            if file_handle is None and hasattr(df, "attrs") and "file_position" in df.attrs:
                                file_position = df.attrs["file_position"]
                            self.read_line_counter += len(df)
                            line_counter = self.read_line_counter
                        except StopIteration:
                            self.file_semaphore.release()
                            break
                        self.file_semaphore.release()
                        yield FileChunkDataFrame(df, file_name, file_index, chunk_index, previous_file_position, line_counter)
                        previous_file_position = file_position

    def load_sample_df(self, resources_base_dir:str, *, upload_alter:bool=True, file_index:int=0, allow_chunks:bool=True, **kwargs) -> GeneralDataFrame:
        generator = self.get_local_df_chunk_generator(resources_base_dir=resources_base_dir, ckan=None, allow_chunks=allow_chunks)
        chunk = next(generator)
        df_local = chunk.df
        generator.close()
        if upload_alter:
            df_upload = self.df_mapper.df_upload_alter(df_local, total_lines_read=chunk.read_line_counter,
                                                       fields=self._get_fields_info(), file_query=chunk.file_path)
            return df_upload
        else:
            return df_local

    def upload_file_chunk(self, ckan:CkanApi, package_id:str, file_chunk:FileChunkDataFrame, *,
                          reupload:bool=False, override_ckan:bool=False, cancel_if_present:bool=True) -> CkanResourceInfo:
        file_path = file_chunk.file_path
        ds_builder, file_dir = self._data_store_builder_of_file(file_path=file_path)
        if file_chunk.is_first_chunk:
            return ds_builder.patch_request(ckan=ckan, package_id=package_id, reupload=reupload,
                                            resources_base_dir=file_dir)
        else:
            ds_builder.upsert_request_df(ckan, file_chunk.df, file_name=file_path, total_lines_read=self.read_line_counter)
            return ckan.map.get_resource_info(resource_name=ds_builder.name, package_name=package_id)


    ## Download --------------
    def download_file_query_item_df(self, ckan: CkanApi, out_dir: str, file_query_item: str, full_download:bool=True) -> Tuple[str, pd.DataFrame]:
        resource_name = file_query_item
        ds_builder, _ = self._data_store_builder_of_file(file_path=resource_name)
        file_dir = resolve_rel_path(out_dir, glob_rm_glob(self.dir_name), field=f"File/URL of resource {self.name}")
        df = ds_builder.download_request(ckan, out_dir=file_dir, full_download=full_download)
        if df is not None:
            self.read_line_counter += len(df)
        return ds_builder.downloaded_destination, df

    def download_file_query_item(self, ckan: CkanApi, out_dir: str, file_query_item: str, full_download:bool=True) -> Tuple[Union[str,None], Union[requests.Response,None]]:
        downloaded_destination, df = self.download_file_query_item_df(ckan=ckan, out_dir=out_dir, file_query_item=file_query_item,full_download=full_download)
        return downloaded_destination, None

    def download_request_generator_df(self, ckan: CkanApi, out_dir: str,
                                   excluded_resource_names:Set[str]=None) -> Generator[Tuple[Union[str,None], Union[pd.DataFrame,None]], Any, None]:
        self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True,
                                           excluded_resource_names=excluded_resource_names)
        for file_query_item in self.get_file_query_generator():
            yield self.download_file_query_item_df(ckan=ckan, out_dir=out_dir, file_query_item=file_query_item)




#!python3
# -*- coding: utf-8 -*-
"""
Package with helper functions for CKAN requests using pandas DataFrames.
"""

from . import ckan_api_params
from . import ckan_api_0_base
from . import ckan_api_1_map
from . import ckan_api_2_readonly
from . import ckan_api_3_policy
from . import ckan_api_4_readwrite
from . import ckan_api_5_manage
from . import ckan_api
# from . import deprecated

# usage shortcuts
from ckanapi_harvesters.ckan_api.ckan_api import CkanApi, CkanApiParams, CkanApiABC, CKAN_API_VERSION, CkanApiMap


